import enum
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.integrations.logging import (
    BreadcrumbHandler,
    EventHandler,
    _BaseHandler,
)
from sentry_sdk.logger import _log_level_to_otel
from sentry_sdk.utils import safe_repr

if TYPE_CHECKING:
    from logging import LogRecord
    from typing import Any, Optional, Union

try:
    import loguru
    from loguru import logger
    from loguru._defaults import LOGURU_FORMAT as DEFAULT_FORMAT

    if TYPE_CHECKING:
        from loguru import Message
except ImportError:
    raise DidNotEnable("LOGURU is not installed")


_SENTINEL = object()


class LoggingLevels(enum.IntEnum):
    TRACE = 5
    DEBUG = 10
    INFO = 20
    SUCCESS = 25
    WARNING = 30
    ERROR = 40
    CRITICAL = 50


DEFAULT_LEVEL = LoggingLevels.INFO.value
DEFAULT_EVENT_LEVEL = LoggingLevels.ERROR.value


SENTRY_LEVEL_FROM_LOGURU_LEVEL = {
    "TRACE": "DEBUG",
    "DEBUG": "DEBUG",
    "INFO": "INFO",
    "SUCCESS": "INFO",
    "WARNING": "WARNING",
    "ERROR": "ERROR",
    "CRITICAL": "CRITICAL",
}

# Map Loguru level numbers to corresponding OTel level numbers
SEVERITY_TO_OTEL_SEVERITY = {
    LoggingLevels.CRITICAL: 21,  # fatal
    LoggingLevels.ERROR: 17,  # error
    LoggingLevels.WARNING: 13,  # warn
    LoggingLevels.SUCCESS: 11,  # info
    LoggingLevels.INFO: 9,  # info
    LoggingLevels.DEBUG: 5,  # debug
    LoggingLevels.TRACE: 1,  # trace
}


class LoguruIntegration(Integration):
    identifier = "loguru"

    level: "Optional[int]" = DEFAULT_LEVEL
    event_level: "Optional[int]" = DEFAULT_EVENT_LEVEL
    breadcrumb_format = DEFAULT_FORMAT
    event_format = DEFAULT_FORMAT
    sentry_logs_level: "Optional[int]" = DEFAULT_LEVEL
    capture_sentry_logs: "Optional[Union[bool, object]]" = _SENTINEL

    def __init__(
        self,
        level: "Optional[int]" = DEFAULT_LEVEL,
        event_level: "Optional[int]" = DEFAULT_EVENT_LEVEL,
        breadcrumb_format: "str | loguru.FormatFunction" = DEFAULT_FORMAT,
        event_format: "str | loguru.FormatFunction" = DEFAULT_FORMAT,
        sentry_logs_level: "Optional[int]" = DEFAULT_LEVEL,
        capture_sentry_logs: "Optional[Union[bool, object]]" = _SENTINEL,
    ) -> None:
        LoguruIntegration.level = level
        LoguruIntegration.event_level = event_level
        LoguruIntegration.breadcrumb_format = breadcrumb_format
        LoguruIntegration.event_format = event_format
        LoguruIntegration.sentry_logs_level = sentry_logs_level
        LoguruIntegration.capture_sentry_logs = capture_sentry_logs

    @staticmethod
    def setup_once() -> None:
        if LoguruIntegration.level is not None:
            logger.add(
                LoguruBreadcrumbHandler(level=LoguruIntegration.level),
                level=LoguruIntegration.level,
                format=LoguruIntegration.breadcrumb_format,
            )

        if LoguruIntegration.event_level is not None:
            logger.add(
                LoguruEventHandler(level=LoguruIntegration.event_level),
                level=LoguruIntegration.event_level,
                format=LoguruIntegration.event_format,
            )

        if LoguruIntegration.sentry_logs_level is not None:
            logger.add(
                loguru_sentry_logs_handler,
                level=LoguruIntegration.sentry_logs_level,
            )


class _LoguruBaseHandler(_BaseHandler):
    def __init__(self, *args: "Any", **kwargs: "Any") -> None:
        if kwargs.get("level"):
            kwargs["level"] = SENTRY_LEVEL_FROM_LOGURU_LEVEL.get(
                kwargs.get("level", ""), DEFAULT_LEVEL
            )

        super().__init__(*args, **kwargs)

    def _logging_to_event_level(self, record: "LogRecord") -> str:
        try:
            return SENTRY_LEVEL_FROM_LOGURU_LEVEL[
                LoggingLevels(record.levelno).name
            ].lower()
        except (ValueError, KeyError):
            return record.levelname.lower() if record.levelname else ""


class LoguruEventHandler(_LoguruBaseHandler, EventHandler):
    """Modified version of :class:`sentry_sdk.integrations.logging.EventHandler` to use loguru's level names."""

    pass


class LoguruBreadcrumbHandler(_LoguruBaseHandler, BreadcrumbHandler):
    """Modified version of :class:`sentry_sdk.integrations.logging.BreadcrumbHandler` to use loguru's level names."""

    pass


def loguru_sentry_logs_handler(message: "Message") -> None:
    # This is intentionally a callable sink instead of a standard logging handler
    # since otherwise we wouldn't get direct access to message.record
    client = sentry_sdk.get_client()
    if not client.is_active():
        return

    # TODO: remove this compat hack in the next major. Capture should
    # only depend on capture_sentry_logs being True.
    compat_logs_enabled = client.options.get("enable_logs", False) or client.options[
        "_experiments"
    ].get("enable_logs", False)
    should_capture_logs = False
    if LoguruIntegration.capture_sentry_logs is True:
        should_capture_logs = True
    elif LoguruIntegration.capture_sentry_logs is _SENTINEL and compat_logs_enabled:
        should_capture_logs = True

    if not should_capture_logs:
        return

    record = message.record

    if (
        LoguruIntegration.sentry_logs_level is None
        or record["level"].no < LoguruIntegration.sentry_logs_level
    ):
        return

    otel_severity_number, otel_severity_text = _log_level_to_otel(
        record["level"].no, SEVERITY_TO_OTEL_SEVERITY
    )

    attrs: "dict[str, Any]" = {"sentry.origin": "auto.log.loguru"}

    project_root = client.options["project_root"]
    if record.get("file"):
        if project_root is not None and record["file"].path.startswith(project_root):
            attrs["code.file.path"] = record["file"].path[len(project_root) + 1 :]
        else:
            attrs["code.file.path"] = record["file"].path

    if record.get("line") is not None:
        attrs["code.line.number"] = record["line"]

    if record.get("function"):
        attrs["code.function.name"] = record["function"]

    if record.get("thread"):
        attrs["thread.name"] = record["thread"].name
        attrs["thread.id"] = record["thread"].id

    if record.get("process"):
        attrs["process.pid"] = record["process"].id
        attrs["process.executable.name"] = record["process"].name

    if record.get("name"):
        attrs["logger.name"] = record["name"]

    extra = record.get("extra")
    if isinstance(extra, dict):
        for key, value in extra.items():
            if isinstance(value, (str, int, float, bool)):
                attrs[f"sentry.message.parameter.{key}"] = value
            else:
                attrs[f"sentry.message.parameter.{key}"] = safe_repr(value)

    sentry_sdk.get_current_scope()._capture_log(
        {
            "severity_text": otel_severity_text,
            "severity_number": otel_severity_number,
            "body": record["message"],
            "attributes": attrs,
            "time_unix_nano": int(record["time"].timestamp() * 1e9),
            "trace_id": None,
            "span_id": None,
        }
    )



from functools import wraps
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import SPANDATA
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.utils import has_data_collection_enabled

from ..spans import execute_tool_span, update_execute_tool_span

if TYPE_CHECKING:
    from typing import Any, Awaitable, Callable

try:
    import agents
except ImportError:
    raise DidNotEnable("OpenAI Agents not installed")


async def _get_all_tools(
    original_get_all_tools: "Callable[..., Awaitable[list[agents.Tool]]]",
    agent: "agents.Agent",
    context_wrapper: "agents.RunContextWrapper",
) -> "list[agents.Tool]":
    """
    Responsible for creating and managing `gen_ai.execute_tool` spans.
    """
    # Get the original tools
    tools = await original_get_all_tools(agent, context_wrapper)

    wrapped_tools = []
    for tool in tools:
        # Wrap only the function tools (for now)
        if tool.__class__.__name__ != "FunctionTool":
            wrapped_tools.append(tool)
            continue

        # Create a new FunctionTool with our wrapped invoke method
        original_on_invoke = tool.on_invoke_tool

        def create_wrapped_invoke(
            current_tool: "agents.Tool", current_on_invoke: "Callable[..., Any]"
        ) -> "Callable[..., Any]":
            @wraps(current_on_invoke)
            async def sentry_wrapped_on_invoke_tool(
                *args: "Any", **kwargs: "Any"
            ) -> "Any":
                with execute_tool_span(current_tool, *args, **kwargs) as span:
                    # We can not capture exceptions in tool execution here because
                    # `_on_invoke_tool` is swallowing the exception here:
                    # https://github.com/openai/openai-agents-python/blob/main/src/agents/tool.py#L409-L422
                    # And because function_tool is a decorator with `default_tool_error_function` set as a default parameter
                    # I was unable to monkey patch it because those are evaluated at module import time
                    # and the SDK is too late to patch it. I was also unable to patch `_on_invoke_tool_impl`
                    # because it is nested inside this import time code. As if they made it hard to patch on purpose...
                    result = await current_on_invoke(*args, **kwargs)
                    update_execute_tool_span(span, agent, current_tool, result)

                    client = sentry_sdk.get_client()
                    if has_data_collection_enabled(client.options):
                        if not client.options["data_collection"]["gen_ai"]["inputs"]:
                            return result
                    elif not should_send_default_pii():
                        return result

                    if isinstance(span, StreamedSpan):
                        span.set_attribute(SPANDATA.GEN_AI_TOOL_INPUT, args[1])
                    else:
                        span.set_data(SPANDATA.GEN_AI_TOOL_INPUT, args[1])

                return result

            return sentry_wrapped_on_invoke_tool

        tool.on_invoke_tool = create_wrapped_invoke(tool, original_on_invoke)
        wrapped_tools.append(tool)

    return wrapped_tools



"""
Data Collection configuration.

Implements the ``data_collection`` client option described in the Sentry SDK
"Data Collection" spec
(https://develop.sentry.dev/sdk/foundations/client/data-collection/).

``data_collection`` supersedes the single ``send_default_pii`` boolean with a
structured configuration that lets users enable or restrict automatically
collected data by category (user identity, cookies, HTTP headers, URL query params,
HTTP bodies, generative AI inputs/outputs, stack frame variables, source
context).

Resolution precedence (see :func:`_resolve_data_collection`):

* ``data_collection`` set, ``send_default_pii`` unset -> honour ``data_collection``
  using the spec defaults for any omitted field.
* ``send_default_pii`` set, ``data_collection`` unset -> derive a
  resolved ``DataCollection`` that mirrors what ``send_default_pii`` collects today.
* neither set -> treated as ``send_default_pii=False``.
* both set -> ``data_collection`` wins (it is the single source of truth); a
  ``DeprecationWarning`` is emitted for ``send_default_pii``.
"""

import warnings
from typing import TYPE_CHECKING, List, Mapping, Optional, Union, cast
from urllib.parse import parse_qs, urlencode

from sentry_sdk._types import SENSITIVE_DATA_SUBSTITUTE

if TYPE_CHECKING:
    from typing import Any, Dict

    from sentry_sdk._types import (
        DataCollection,
        GenAICollectionBehaviour,
        GraphQLCollectionBehaviour,
        HttpHeadersCollectionBehaviour,
        KeyValueCollectionBehaviour,
    )

# ``http_bodies`` defaults to this (collect everything the
# platform supports); an empty list is the explicit opt-out.
# response bodyies are not included here because we don't
# currently capture them (as of Jul 7 2026)
_ALL_HTTP_BODY_TYPES = [
    "incoming_request",
    "outgoing_request",
]

# Default number of source lines captured above and below a stack frame.
_DEFAULT_FRAME_CONTEXT_LINES = 5

# Collection modes for key-value data (cookies, headers, URL query params).
# snake_case (Python-only deviation from the spec's camelCase); never
# serialized to Sentry.
_VALID_KEY_VALUE_COLLECTION_BEHAVIOUR_MODES = ("off", "denylist", "allowlist")

# Values of keys that contain any of
# these terms (partial, case-insensitive) are always replaced with
# ``"[Filtered]"`` regardless of the configured collection mode.
_SENSITIVE_DENYLIST = [
    "auth",
    "token",
    "secret",
    "password",
    "passwd",
    "pwd",
    "key",
    "jwt",
    "bearer",
    "sso",
    "saml",
    "csrf",
    "xsrf",
    "credentials",
    "session",
    "sid",
    "identity",
]


def _is_sensitive_key(key: str, extra_terms: "Optional[List[str]]" = None) -> bool:
    """
    Return whether ``key`` matches the sensitive denylist using a partial,
    case-insensitive substring match.

    :param extra_terms: additional deny terms (e.g. user-provided) to consider
        alongside the built-in `_SENSITIVE_DENYLIST`.
    """
    lowered = key.lower()
    for term in _SENSITIVE_DENYLIST:
        if term in lowered:
            return True
    if extra_terms:
        for term in extra_terms:
            if term and term.lower() in lowered:
                return True
    return False


def _apply_data_collection_filtering_to_query_string(
    query_string: str,
    behaviour: "KeyValueCollectionBehaviour",
) -> "Union[str, None]":
    parsed_qs = parse_qs(query_string, keep_blank_values=True)
    filtered_qs = _apply_key_value_collection_filtering(
        items=parsed_qs, behaviour=behaviour
    )

    if filtered_qs:
        return urlencode(filtered_qs, doseq=True)

    return None


def _apply_key_value_collection_filtering(
    items: "Mapping[str, Any]",
    behaviour: "KeyValueCollectionBehaviour",
    substitute: "Any" = SENSITIVE_DATA_SUBSTITUTE,
) -> "Dict[str, Any]":

    if behaviour["mode"] == "off":
        return {}

    result: "Dict[str, Any]" = {}

    if behaviour["mode"] == "allowlist":
        for key, value in items.items():
            is_allowed = False
            if isinstance(key, str):
                lowered = key.lower()
                is_allowed = any(
                    term and term.lower() in lowered
                    for term in behaviour.get("terms", [])
                )
            if is_allowed and not _is_sensitive_key(key):
                result[key] = value
            else:
                result[key] = substitute
        return result

    # denylist behaviour
    for key, value in items.items():
        if isinstance(key, str) and _is_sensitive_key(
            key=key, extra_terms=behaviour.get("terms", [])
        ):
            result[key] = substitute
        else:
            result[key] = value
    return result


def _map_from_send_default_pii(
    *,
    send_default_pii: bool,
    include_local_variables: bool,
    include_source_context: bool,
) -> "DataCollection":
    """
    Build a fully-resolved ``DataCollection`` dict that mirrors the data
    ``send_default_pii`` collects today. Used when ``data_collection`` is not
    provided explicitly.
    """
    terms = [] if send_default_pii else ["forwarded", "-ip", "remote-", "via", "-user"]

    return {
        "provided_by_user": False,
        "user_info": send_default_pii,
        "cookies": {"mode": "denylist", "terms": terms},
        # Headers are collected in both PII modes today (sensitive ones filtered
        # when PII is off), so this never maps to "off".
        "http_headers": {
            "request": {"mode": "denylist", "terms": terms},
        },
        # Bodies are collected regardless of PII today, bounded by
        # ``max_request_body_size``.
        "http_bodies": list(_ALL_HTTP_BODY_TYPES),
        "url_query_params": {"mode": "denylist", "terms": terms},
        "graphql": {"document": send_default_pii, "variables": send_default_pii},
        "gen_ai": {"inputs": send_default_pii, "outputs": send_default_pii},
        "database_query_data": send_default_pii,
        "queues": send_default_pii,
        "stack_frame_variables": include_local_variables,
        "frame_context_lines": (
            _DEFAULT_FRAME_CONTEXT_LINES if include_source_context else 0
        ),
    }


def _resolve_explicit(
    d: "dict[str, Any]",
    include_local_variables: bool,
    include_source_context: bool,
) -> "DataCollection":
    """
    Build a fully-resolved ``DataCollection`` from a user-supplied
    ``data_collection`` dict, filling in spec defaults for any omitted or
    partially-specified field. Frame fields fall back to the legacy
    ``include_local_variables`` / ``include_source_context`` options when unset.
    """
    # frame_context_lines accepts an integer or a boolean fallback (spec: True
    # -> platform default of 5, False -> 0). bool is a subclass of int, so
    # coerce explicitly before treating it as a line count.
    frame_context_lines = d.get("frame_context_lines")
    if frame_context_lines is None:
        frame_context_lines = (
            _DEFAULT_FRAME_CONTEXT_LINES if include_source_context else 0
        )
    elif isinstance(frame_context_lines, bool):
        frame_context_lines = _DEFAULT_FRAME_CONTEXT_LINES if frame_context_lines else 0

    stack_frame_variables = d.get("stack_frame_variables")
    if stack_frame_variables is None:
        stack_frame_variables = include_local_variables

    # http_bodies: omitted means "all valid types"; [] is the explicit opt-out.
    http_bodies = d.get("http_bodies")
    http_bodies = (
        list(http_bodies) if http_bodies is not None else list(_ALL_HTTP_BODY_TYPES)
    )

    return {
        "provided_by_user": True,
        "user_info": d.get("user_info", True),
        "cookies": _kvcb_from_value(d.get("cookies") or {}),
        "http_headers": _http_headers_from_value(d.get("http_headers") or {}),
        "http_bodies": http_bodies,
        "url_query_params": _kvcb_from_value(d.get("url_query_params") or {}),
        "graphql": _graphql_from_value(d.get("graphql") or {}),
        "gen_ai": _gen_ai_from_value(d.get("gen_ai") or {}),
        "database_query_data": d.get("database_query_data", True),
        "queues": d.get("queues", True),
        "stack_frame_variables": stack_frame_variables,
        "frame_context_lines": frame_context_lines,
    }


def _kvcb_from_value(
    val: "dict[str, Any]",
) -> "KeyValueCollectionBehaviour":
    mode = val.get("mode", "denylist")
    terms = val.get("terms", None)

    if mode not in _VALID_KEY_VALUE_COLLECTION_BEHAVIOUR_MODES:
        raise ValueError(
            "Invalid collection mode {!r}. Must be one of {}.".format(
                mode, _VALID_KEY_VALUE_COLLECTION_BEHAVIOUR_MODES
            )
        )

    behaviour: "dict[str, Any]" = {"mode": mode}
    if terms is not None:
        behaviour["terms"] = list(terms)
    return cast("KeyValueCollectionBehaviour", behaviour)


def _http_headers_from_value(
    val: "dict[str, Any]",
) -> "HttpHeadersCollectionBehaviour":
    return {
        "request": (
            _kvcb_from_value(val["request"])
            if "request" in val
            else _kvcb_from_value({"mode": "denylist"})
        ),
    }


def _gen_ai_from_value(val: "dict[str, Any]") -> "GenAICollectionBehaviour":
    return {
        "inputs": val.get("inputs", True),
        "outputs": val.get("outputs", True),
    }


def _graphql_from_value(
    val: "dict[str, Any]",
) -> "GraphQLCollectionBehaviour":
    return {
        "document": val.get("document", True),
        "variables": val.get("variables", True),
    }


def _resolve_data_collection(options: "Dict[str, Any]") -> "DataCollection":
    """
    Resolve the effective ``DataCollection`` dict from client ``options``.

    Reads ``data_collection``, ``send_default_pii``, ``include_local_variables``
    and ``include_source_context`` and returns a fully-resolved dict with
    concrete values for every field.

    ``data_collection`` must be a plain ``dict``.
    """
    user_dc = options.get("_experiments", {}).get("data_collection")
    send_default_pii = options.get("send_default_pii")

    include_local_variables = (
        bool(options.get("include_local_variables"))
        if options.get("include_local_variables") is not None
        else True
    )
    include_source_context = (
        bool(options.get("include_source_context"))
        if options.get("include_source_context") is not None
        else True
    )

    if user_dc is not None:
        if not isinstance(user_dc, dict):
            raise TypeError(
                "`data_collection` must be a dict, got {!r}.".format(
                    type(user_dc).__name__
                )
            )
        if send_default_pii is not None:
            warnings.warn(
                "`send_default_pii` is deprecated and ignored when "
                "`data_collection` is set.",
                DeprecationWarning,
                stacklevel=2,
            )
        return _resolve_explicit(
            user_dc,
            include_local_variables,
            include_source_context,
        )

    return _map_from_send_default_pii(
        send_default_pii=bool(send_default_pii),
        include_local_variables=include_local_variables,
        include_source_context=include_source_context,
    )



from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.ai.utils import (
    get_start_span_function,
    normalize_message_roles,
    set_data_normalized,
    truncate_and_annotate_messages,
)
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import (
    has_span_streaming_enabled,
    should_truncate_gen_ai_input,
)
from sentry_sdk.utils import has_data_collection_enabled, safe_serialize

from ..consts import SPAN_ORIGIN
from ..utils import _set_agent_data, _set_usage_data

if TYPE_CHECKING:
    from typing import Any, Union

    import agents


def invoke_agent_span(
    context: "agents.RunContextWrapper", agent: "agents.Agent", kwargs: "dict[str, Any]"
) -> "Union[sentry_sdk.tracing.Span, StreamedSpan]":
    client_options = sentry_sdk.get_client().options
    span_streaming = has_span_streaming_enabled(client_options)
    if span_streaming:
        span = sentry_sdk.traces.start_span(
            name=f"invoke_agent {agent.name}",
            attributes={
                "sentry.op": OP.GEN_AI_INVOKE_AGENT,
                "sentry.origin": SPAN_ORIGIN,
                SPANDATA.GEN_AI_OPERATION_NAME: "invoke_agent",
            },
        )
    else:
        start_span_function = get_start_span_function()
        span = start_span_function(
            op=OP.GEN_AI_INVOKE_AGENT,
            name=f"invoke_agent {agent.name}",
            origin=SPAN_ORIGIN,
        )
        span.__enter__()

        span.set_data(SPANDATA.GEN_AI_OPERATION_NAME, "invoke_agent")

    record_inputs = False
    if has_data_collection_enabled(client_options):
        if client_options["data_collection"]["gen_ai"]["inputs"]:
            record_inputs = True
    elif should_send_default_pii():
        record_inputs = True

    if record_inputs:
        messages = []
        if agent.instructions:
            message = (
                agent.instructions
                if isinstance(agent.instructions, str)
                else safe_serialize(agent.instructions)
            )
            messages.append(
                {
                    "content": [{"text": message, "type": "text"}],
                    "role": "system",
                }
            )

        original_input = kwargs.get("original_input")
        if original_input is not None:
            message = (
                original_input
                if isinstance(original_input, str)
                else safe_serialize(original_input)
            )
            messages.append(
                {
                    "content": [{"text": message, "type": "text"}],
                    "role": "user",
                }
            )

        if len(messages) > 0:
            normalized_messages = normalize_message_roles(messages)
            client = sentry_sdk.get_client()
            scope = sentry_sdk.get_current_scope()
            messages_data = (
                truncate_and_annotate_messages(normalized_messages, span, scope)
                if should_truncate_gen_ai_input(client.options)
                else normalized_messages
            )
            if messages_data is not None:
                set_data_normalized(
                    span,
                    SPANDATA.GEN_AI_REQUEST_MESSAGES,
                    messages_data,
                    unpack=False,
                )

    _set_agent_data(span, agent)

    return span


def update_invoke_agent_span(
    span: "Union[sentry_sdk.tracing.Span, StreamedSpan]",
    context: "agents.RunContextWrapper",
    agent: "agents.Agent",
    output: "Any" = None,
) -> None:
    # Add aggregated usage data from context_wrapper
    if hasattr(context, "usage"):
        _set_usage_data(span, context.usage)

    client = sentry_sdk.get_client()
    if has_data_collection_enabled(client.options):
        if client.options["data_collection"]["gen_ai"]["outputs"]:
            set_data_normalized(
                span, SPANDATA.GEN_AI_RESPONSE_TEXT, output, unpack=False
            )
    elif should_send_default_pii():
        set_data_normalized(span, SPANDATA.GEN_AI_RESPONSE_TEXT, output, unpack=False)

    # Add conversation ID from agent
    conv_id = getattr(agent, "_sentry_conversation_id", None)
    if conv_id:
        if isinstance(span, StreamedSpan):
            span.set_attribute(SPANDATA.GEN_AI_CONVERSATION_ID, conv_id)
        else:
            span.set_data(SPANDATA.GEN_AI_CONVERSATION_ID, conv_id)



import sys
from functools import wraps
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.integrations import DidNotEnable
from sentry_sdk.utils import capture_internal_exceptions, reraise

from ..spans import execute_tool_span, update_execute_tool_span
from ..utils import _capture_exception, get_current_agent

if TYPE_CHECKING:
    from typing import Any

try:
    try:
        from pydantic_ai.tool_manager import ToolManager
    except ImportError:
        from pydantic_ai._tool_manager import ToolManager  # type: ignore

    from pydantic_ai.exceptions import ToolRetryError
except ImportError:
    raise DidNotEnable("pydantic-ai not installed")


def _patch_tool_execution() -> None:
    if hasattr(ToolManager, "execute_tool_call"):
        _patch_execute_tool_call()

    elif hasattr(ToolManager, "_call_tool"):
        # older versions
        _patch_call_tool()


def _patch_execute_tool_call() -> None:
    original_execute_tool_call = ToolManager.execute_tool_call

    @wraps(original_execute_tool_call)
    async def wrapped_execute_tool_call(
        self: "Any", validated: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        if not validated or not hasattr(validated, "call"):
            return await original_execute_tool_call(self, validated, *args, **kwargs)

        # Extract tool info before calling original
        call = validated.call
        name = call.tool_name
        tool = self.tools.get(name) if self.tools else None
        selected_tool_definition = getattr(tool, "tool_def", None)

        # Get agent from contextvar
        agent = get_current_agent()

        if agent and tool:
            try:
                args_dict = call.args_as_dict()
            except Exception:
                args_dict = call.args if isinstance(call.args, dict) else {}

            # Create execute_tool span
            # Nesting is handled by isolation_scope() to ensure proper parent-child relationships
            with sentry_sdk.isolation_scope():
                with execute_tool_span(
                    name,
                    args_dict,
                    agent,
                    tool_definition=selected_tool_definition,
                ) as span:
                    try:
                        result = await original_execute_tool_call(
                            self,
                            validated,
                            *args,
                            **kwargs,
                        )
                        update_execute_tool_span(span, result)
                        return result
                    except ToolRetryError as exc:
                        exc_info = sys.exc_info()
                        with capture_internal_exceptions():
                            # Avoid circular import due to multi-file integration structure
                            from sentry_sdk.integrations.pydantic_ai import (
                                PydanticAIIntegration,
                            )

                            integration = sentry_sdk.get_client().get_integration(
                                PydanticAIIntegration
                            )
                            if (
                                integration is not None
                                and integration.handled_tool_call_exceptions
                            ):
                                _capture_exception(exc, handled=True)
                        reraise(*exc_info)

        return await original_execute_tool_call(self, validated, *args, **kwargs)

    ToolManager.execute_tool_call = wrapped_execute_tool_call  # type: ignore[method-assign]


def _patch_call_tool() -> None:
    """
    Patch ToolManager._call_tool to create execute_tool spans.

    This is the single point where ALL tool calls flow through in pydantic_ai,
    regardless of toolset type (function, MCP, combined, wrapper, etc.).

    By patching here, we avoid:
    - Patching multiple toolset classes
    - Dealing with signature mismatches from instrumented MCP servers
    - Complex nested toolset handling
    """
    original_call_tool = ToolManager._call_tool  # type: ignore[attr-defined]

    @wraps(original_call_tool)
    async def wrapped_call_tool(
        self: "Any", call: "Any", *args: "Any", **kwargs: "Any"
    ) -> "Any":
        # Extract tool info before calling original
        name = call.tool_name
        tool = self.tools.get(name) if self.tools else None
        selected_tool_definition = getattr(tool, "tool_def", None)

        # Get agent from contextvar
        agent = get_current_agent()

        if agent and tool:
            try:
                args_dict = call.args_as_dict()
            except Exception:
                args_dict = call.args if isinstance(call.args, dict) else {}

            # Create execute_tool span
            # Nesting is handled by isolation_scope() to ensure proper parent-child relationships
            with sentry_sdk.isolation_scope():
                with execute_tool_span(
                    name,
                    args_dict,
                    agent,
                    tool_definition=selected_tool_definition,
                ) as span:
                    try:
                        result = await original_call_tool(
                            self,
                            call,
                            *args,
                            **kwargs,
                        )
                        update_execute_tool_span(span, result)
                        return result
                    except ToolRetryError as exc:
                        exc_info = sys.exc_info()
                        with capture_internal_exceptions():
                            # Avoid circular import due to multi-file integration structure
                            from sentry_sdk.integrations.pydantic_ai import (
                                PydanticAIIntegration,
                            )

                            integration = sentry_sdk.get_client().get_integration(
                                PydanticAIIntegration
                            )
                            if (
                                integration is not None
                                and integration.handled_tool_call_exceptions
                            ):
                                _capture_exception(exc, handled=True)
                        reraise(*exc_info)

        # No span context - just call original
        return await original_call_tool(
            self,
            call,
            *args,
            **kwargs,
        )

    ToolManager._call_tool = wrapped_call_tool  # type: ignore[attr-defined]



from __future__ import annotations

import contextlib
import re
from typing import Any, Awaitable, Callable, Iterator, TypeVar, Union

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing import Span
from sentry_sdk.tracing_utils import (
    add_query_source,
    has_span_streaming_enabled,
    record_sql_queries,
)
from sentry_sdk.utils import (
    capture_internal_exceptions,
    parse_version,
)

try:
    import asyncpg  # type: ignore
    from asyncpg.cursor import (  # type: ignore
        BaseCursor,
        Cursor,
        CursorIterator,
    )

except ImportError:
    raise DidNotEnable("asyncpg not installed.")


class AsyncPGIntegration(Integration):
    identifier = "asyncpg"
    origin = f"auto.db.{identifier}"
    _record_params = False

    def __init__(self, *, record_params: bool = False):
        AsyncPGIntegration._record_params = record_params

    @staticmethod
    def setup_once() -> None:
        # asyncpg.__version__ is a string containing the semantic version in the form of "<major>.<minor>.<patch>"
        asyncpg_version = parse_version(asyncpg.__version__)
        _check_minimum_version(AsyncPGIntegration, asyncpg_version)

        asyncpg.Connection.execute = _wrap_execute(
            asyncpg.Connection.execute,
        )

        asyncpg.Connection._execute = _wrap_connection_method(
            asyncpg.Connection._execute
        )
        asyncpg.Connection._executemany = _wrap_connection_method(
            asyncpg.Connection._executemany, executemany=True
        )
        asyncpg.Connection.prepare = _wrap_connection_method(asyncpg.Connection.prepare)

        BaseCursor._bind_exec = _wrap_cursor_method(BaseCursor._bind_exec)
        BaseCursor._exec = _wrap_cursor_method(BaseCursor._exec)

        asyncpg.connect_utils._connect_addr = _wrap_connect_addr(
            asyncpg.connect_utils._connect_addr
        )


T = TypeVar("T")


def _normalize_query(query: str) -> str:
    return re.sub(r"\s+", " ", query).strip()


def _wrap_execute(f: "Callable[..., Awaitable[T]]") -> "Callable[..., Awaitable[T]]":
    async def _inner(*args: "Any", **kwargs: "Any") -> "T":
        client = sentry_sdk.get_client()
        if client.get_integration(AsyncPGIntegration) is None:
            return await f(*args, **kwargs)

        # Avoid recording calls to _execute twice.
        # Calls to Connection.execute with args also call
        # Connection._execute, which is recorded separately
        # args[0] = the connection object, args[1] is the query
        if len(args) > 2:
            return await f(*args, **kwargs)

        query = _normalize_query(args[1])
        with record_sql_queries(
            cursor=None,
            query=query,
            params_list=None,
            paramstyle=None,
            executemany=False,
            span_origin=AsyncPGIntegration.origin,
        ) as span:
            res = await f(*args, **kwargs)
            if isinstance(span, StreamedSpan):
                with capture_internal_exceptions():
                    add_query_source(span)

        if not isinstance(span, StreamedSpan):
            with capture_internal_exceptions():
                add_query_source(span)

        return res

    return _inner


SubCursor = TypeVar("SubCursor", bound=BaseCursor)


@contextlib.contextmanager
def _record(
    cursor: "SubCursor | None",
    query: str,
    params_list: "tuple[Any, ...] | None",
    *,
    executemany: bool = False,
) -> "Iterator[Union[Span, StreamedSpan]]":
    client = sentry_sdk.get_client()
    integration = client.get_integration(AsyncPGIntegration)
    if integration is not None and not integration._record_params:
        params_list = None

    param_style = "pyformat" if params_list else None

    query = _normalize_query(query)
    with record_sql_queries(
        cursor=cursor,
        query=query,
        params_list=params_list,
        paramstyle=param_style,
        executemany=executemany,
        record_cursor_repr=cursor is not None,
        span_origin=AsyncPGIntegration.origin,
    ) as span:
        yield span


def _wrap_connection_method(
    f: "Callable[..., Awaitable[T]]", *, executemany: bool = False
) -> "Callable[..., Awaitable[T]]":
    async def _inner(*args: "Any", **kwargs: "Any") -> "T":
        if sentry_sdk.get_client().get_integration(AsyncPGIntegration) is None:
            return await f(*args, **kwargs)
        query = args[1]
        params_list = args[2] if len(args) > 2 else None
        with _record(None, query, params_list, executemany=executemany) as span:
            _set_db_data(span, args[0])

            res = await f(*args, **kwargs)

            if isinstance(span, StreamedSpan):
                with capture_internal_exceptions():
                    add_query_source(span)

        if not isinstance(span, StreamedSpan):
            with capture_internal_exceptions():
                add_query_source(span)

        return res

    return _inner


def _wrap_cursor_method(
    f: "Callable[..., Awaitable[T]]",
) -> "Callable[..., Awaitable[T]]":
    async def _inner(*args: "Any", **kwargs: "Any") -> "T":
        if sentry_sdk.get_client().get_integration(AsyncPGIntegration) is None:
            return await f(*args, **kwargs)

        cursor = args[0]
        if type(cursor) is CursorIterator:
            span_op_override_value = OP.DB_CURSOR_ITERATOR
        elif type(cursor) is Cursor:
            span_op_override_value = OP.DB_CURSOR_FETCH
        else:
            span_op_override_value = None

        query = _normalize_query(cursor._query)
        with record_sql_queries(
            cursor=cursor,
            query=query,
            params_list=None,
            paramstyle=None,
            executemany=False,
            record_cursor_repr=True,
            span_origin=AsyncPGIntegration.origin,
            span_op_override_value=span_op_override_value,
        ) as span:
            _set_db_data(span, cursor._connection)
            res = await f(*args, **kwargs)

            if isinstance(span, StreamedSpan):
                with capture_internal_exceptions():
                    add_query_source(span)

        if not isinstance(span, StreamedSpan):
            with capture_internal_exceptions():
                add_query_source(span)

        return res

    return _inner


def _wrap_connect_addr(
    f: "Callable[..., Awaitable[T]]",
) -> "Callable[..., Awaitable[T]]":
    async def _inner(*args: "Any", **kwargs: "Any") -> "T":
        client = sentry_sdk.get_client()
        if client.get_integration(AsyncPGIntegration) is None:
            return await f(*args, **kwargs)

        user = kwargs["params"].user
        database = kwargs["params"].database
        addr = kwargs.get("addr")

        if has_span_streaming_enabled(client.options):
            span_attributes = {
                "sentry.op": OP.DB,
                "sentry.origin": AsyncPGIntegration.origin,
                SPANDATA.DB_SYSTEM_NAME: "postgresql",
                SPANDATA.DB_USER: user,
                SPANDATA.DB_NAMESPACE: database,
                SPANDATA.DB_DRIVER_NAME: "asyncpg",
            }
            if addr:
                try:
                    span_attributes[SPANDATA.SERVER_ADDRESS] = addr[0]
                    span_attributes[SPANDATA.SERVER_PORT] = addr[1]
                except IndexError:
                    pass

            with capture_internal_exceptions():
                sentry_sdk.add_breadcrumb(
                    message="connect", category="query", data=span_attributes
                )

            if sentry_sdk.traces.get_current_span() is None:
                return await f(*args, **kwargs)

            with sentry_sdk.traces.start_span(
                name="connect", attributes=span_attributes
            ):
                return await f(*args, **kwargs)

        with sentry_sdk.start_span(
            op=OP.DB,
            name="connect",
            origin=AsyncPGIntegration.origin,
        ) as span:
            span.set_data(SPANDATA.DB_SYSTEM, "postgresql")
            if addr:
                try:
                    span.set_data(SPANDATA.SERVER_ADDRESS, addr[0])
                    span.set_data(SPANDATA.SERVER_PORT, addr[1])
                except IndexError:
                    pass
            span.set_data(SPANDATA.DB_NAME, database)
            span.set_data(SPANDATA.DB_USER, user)
            span.set_data(SPANDATA.DB_DRIVER_NAME, "asyncpg")

            with capture_internal_exceptions():
                sentry_sdk.add_breadcrumb(
                    message="connect", category="query", data=span._data
                )
            return await f(*args, **kwargs)

    return _inner


def _set_db_data(span: "Union[Span, StreamedSpan]", conn: "Any") -> None:
    addr = conn._addr
    database = conn._params.database
    user = conn._params.user

    if isinstance(span, StreamedSpan):
        span.set_attribute(SPANDATA.DB_SYSTEM_NAME, "postgresql")
        span.set_attribute(SPANDATA.DB_DRIVER_NAME, "asyncpg")
        if addr:
            try:
                span.set_attribute(SPANDATA.SERVER_ADDRESS, addr[0])
                span.set_attribute(SPANDATA.SERVER_PORT, addr[1])
            except IndexError:
                pass

        if database:
            span.set_attribute(SPANDATA.DB_NAMESPACE, database)

        if user:
            span.set_attribute(SPANDATA.DB_USER, user)
    else:
        # Remove this else block once we've completely migrated to streamed spans
        # The use of deprecated attributes here is to ensure backwards compatibility
        span.set_data(SPANDATA.DB_SYSTEM, "postgresql")
        span.set_data(SPANDATA.DB_DRIVER_NAME, "asyncpg")

        if addr:
            try:
                span.set_data(SPANDATA.SERVER_ADDRESS, addr[0])
                span.set_data(SPANDATA.SERVER_PORT, addr[1])
            except IndexError:
                pass

        if database:
            span.set_data(SPANDATA.DB_NAME, database)

        if user:
            span.set_data(SPANDATA.DB_USER, user)



import copy
import json

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA, SPANSTATUS
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import SpanStatus, StreamedSpan
from sentry_sdk.tracing import Span
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import capture_internal_exceptions, has_data_collection_enabled

try:
    from pymongo import monitoring
except ImportError:
    raise DidNotEnable("Pymongo not installed")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, Union

    from pymongo.monitoring import (
        CommandFailedEvent,
        CommandStartedEvent,
        CommandSucceededEvent,
    )


SAFE_COMMAND_ATTRIBUTES = [
    "insert",
    "ordered",
    "find",
    "limit",
    "singleBatch",
    "aggregate",
    "createIndexes",
    "indexes",
    "delete",
    "findAndModify",
    "renameCollection",
    "to",
    "drop",
]


def _strip_pii(command: "Dict[str, Any]") -> "Dict[str, Any]":
    for key in command:
        is_safe_field = key in SAFE_COMMAND_ATTRIBUTES
        if is_safe_field:
            # Skip if safe key
            continue

        update_db_command = key == "update" and "findAndModify" not in command
        if update_db_command:
            # Also skip "update" db command because it is save.
            # There is also an "update" key in the "findAndModify" command, which is NOT safe!
            continue

        # Special stripping for documents
        is_document = key == "documents"
        if is_document:
            for doc in command[key]:
                for doc_key in doc:
                    doc[doc_key] = "%s"
            continue

        # Special stripping for dict style fields
        is_dict_field = key in ["filter", "query", "update"]
        if is_dict_field:
            for item_key in command[key]:
                command[key][item_key] = "%s"
            continue

        # For pipeline fields strip the `$match` dict
        is_pipeline_field = key == "pipeline"
        if is_pipeline_field:
            for pipeline in command[key]:
                for match_key in pipeline["$match"] if "$match" in pipeline else []:
                    pipeline["$match"][match_key] = "%s"
            continue

        # Default stripping
        command[key] = "%s"

    return command


def _get_db_data(event: "Any") -> "Dict[str, Any]":
    data = {}

    client = sentry_sdk.get_client()
    is_span_streaming_enabled = has_span_streaming_enabled(client.options)

    data[SPANDATA.DB_DRIVER_NAME] = "pymongo"
    db_name = event.database_name

    server_address = event.connection_id[0]
    if server_address is not None:
        data[SPANDATA.SERVER_ADDRESS] = server_address

    server_port = event.connection_id[1]
    if server_port is not None:
        data[SPANDATA.SERVER_PORT] = server_port

    if is_span_streaming_enabled:
        data["db.system.name"] = "mongodb"

        if db_name is not None:
            data["db.namespace"] = db_name
    else:
        data[SPANDATA.DB_SYSTEM] = "mongodb"

        if db_name is not None:
            data[SPANDATA.DB_NAME] = db_name

    return data


class CommandTracer(monitoring.CommandListener):
    def __init__(self) -> None:
        self._ongoing_operations: "Dict[int, Union[Span, StreamedSpan]]" = {}

    def _operation_key(
        self,
        event: "Union[CommandFailedEvent, CommandStartedEvent, CommandSucceededEvent]",
    ) -> int:
        return event.request_id

    def started(self, event: "CommandStartedEvent") -> None:
        client = sentry_sdk.get_client()
        if client.get_integration(PyMongoIntegration) is None:
            return

        with capture_internal_exceptions():
            command = dict(copy.deepcopy(event.command))

            command.pop("$db", None)
            command.pop("$clusterTime", None)
            command.pop("$signature", None)

            db_data = _get_db_data(event)

            collection_name = command.get(event.command_name)
            operation_name = event.command_name
            db_name = event.database_name

            lsid = command.pop("lsid", None)
            if has_data_collection_enabled(client.options):
                if not client.options["data_collection"]["database_query_data"]:
                    command = _strip_pii(command)
            elif not should_send_default_pii():
                command = _strip_pii(command)

            query = json.dumps(command, default=str)

            if has_span_streaming_enabled(client.options):
                span_first_data = {
                    "db.operation.name": operation_name,
                    "db.collection.name": collection_name,
                    SPANDATA.DB_QUERY_TEXT: query,
                    "sentry.op": OP.DB,
                    "sentry.origin": PyMongoIntegration.origin,
                    **db_data,
                }

                with capture_internal_exceptions():
                    sentry_sdk.add_breadcrumb(
                        message=query,
                        category="query",
                        type=OP.DB,
                        data=span_first_data,
                    )

                if sentry_sdk.traces.get_current_span() is None:
                    return

                span = sentry_sdk.traces.start_span(
                    name=query, attributes=span_first_data
                )

            else:
                tags = {
                    "db.name": db_name,
                    SPANDATA.DB_SYSTEM: "mongodb",
                    SPANDATA.DB_DRIVER_NAME: "pymongo",
                    SPANDATA.DB_OPERATION: operation_name,
                    # The below is a deprecated field, but leaving for legacy reasons.
                    # The v2 spans will use `db.collection.name` instead.
                    SPANDATA.DB_MONGODB_COLLECTION: collection_name,
                }

                try:
                    tags["net.peer.name"] = event.connection_id[0]
                    tags["net.peer.port"] = str(event.connection_id[1])
                except TypeError:
                    pass

                data: "Dict[str, Any]" = {"operation_ids": {}}
                data["operation_ids"]["operation"] = event.operation_id
                data["operation_ids"]["request"] = event.request_id

                data.update(db_data)

                try:
                    if lsid:
                        lsid_id = lsid["id"]
                        data["operation_ids"]["session"] = str(lsid_id)
                except KeyError:
                    pass

                span = sentry_sdk.start_span(
                    op=OP.DB,
                    name=query,
                    origin=PyMongoIntegration.origin,
                )

                for tag, value in tags.items():
                    # set the tag for backwards-compatibility.
                    # TODO: remove the set_tag call in the next major release!
                    span.set_tag(tag, value)
                    span.set_data(tag, value)

                for key, value in data.items():
                    span.set_data(key, value)

                with capture_internal_exceptions():
                    sentry_sdk.add_breadcrumb(
                        message=query, category="query", type=OP.DB, data=tags
                    )

            self._ongoing_operations[self._operation_key(event)] = span.__enter__()

    def failed(self, event: "CommandFailedEvent") -> None:
        if sentry_sdk.get_client().get_integration(PyMongoIntegration) is None:
            return

        try:
            span = self._ongoing_operations.pop(self._operation_key(event))
            # Ignoring NoOpStreamedSpan as it will always have a status of "ok"
            if type(span) is StreamedSpan:
                span.status = SpanStatus.ERROR
            elif type(span) is Span:
                span.set_status(SPANSTATUS.INTERNAL_ERROR)
            span.__exit__(None, None, None)
        except KeyError:
            return

    def succeeded(self, event: "CommandSucceededEvent") -> None:
        if sentry_sdk.get_client().get_integration(PyMongoIntegration) is None:
            return

        try:
            span = self._ongoing_operations.pop(self._operation_key(event))
            if type(span) is Span:
                span.set_status(SPANSTATUS.OK)
            span.__exit__(None, None, None)
        except KeyError:
            pass


class PyMongoIntegration(Integration):
    identifier = "pymongo"
    origin = f"auto.db.{identifier}"

    @staticmethod
    def setup_once() -> None:
        monitoring.register(CommandTracer())



import math
import sys
from array import array
from collections.abc import Mapping
from datetime import datetime
from itertools import islice
from typing import TYPE_CHECKING

from sentry_sdk.utils import (
    AnnotatedValue,
    capture_internal_exception,
    disable_capture_event,
    format_timestamp,
    safe_repr,
    strip_string,
)

if TYPE_CHECKING:
    from types import TracebackType
    from typing import Any, Callable, ContextManager, Dict, List, Optional, Type, Union

    from sentry_sdk._types import NotImplementedType

    Span = Dict[str, Any]

    ReprProcessor = Callable[[Any, Dict[str, Any]], Union[NotImplementedType, str]]
    Segment = Union[str, int]


# Bytes are technically not strings in Python 3, but we can serialize them
serializable_str_types = (str, bytes, bytearray, memoryview)


# Maximum length of JSON-serialized event payloads that can be safely sent
# before the server may reject the event due to its size. This is not intended
# to reflect actual values defined server-side, but rather only be an upper
# bound for events sent by the SDK.
#
# Can be overwritten if wanting to send more bytes, e.g. with a custom server.
# When changing this, keep in mind that events may be a little bit larger than
# this value due to attached metadata, so keep the number conservative.
MAX_EVENT_BYTES = 10**6

# Maximum depth and breadth of databags. Excess data will be trimmed. If
# max_request_body_size is "always", request bodies won't be trimmed.
MAX_DATABAG_DEPTH = 5
MAX_DATABAG_BREADTH = 10
CYCLE_MARKER = "<cyclic>"


global_repr_processors: "List[ReprProcessor]" = []


def add_global_repr_processor(processor: "ReprProcessor") -> None:
    global_repr_processors.append(processor)


sequence_types: "List[type]" = [tuple, list, set, frozenset, array]


def add_repr_sequence_type(ty: type) -> None:
    sequence_types.append(ty)


class Memo:
    __slots__ = ("_ids", "_objs")

    def __init__(self) -> None:
        self._ids: "Dict[int, Any]" = {}
        self._objs: "List[Any]" = []

    def memoize(self, obj: "Any") -> "ContextManager[bool]":
        self._objs.append(obj)
        return self

    def __enter__(self) -> bool:
        obj = self._objs[-1]
        if id(obj) in self._ids:
            return True
        else:
            self._ids[id(obj)] = obj
            return False

    def __exit__(
        self,
        ty: "Optional[Type[BaseException]]",
        value: "Optional[BaseException]",
        tb: "Optional[TracebackType]",
    ) -> None:
        self._ids.pop(id(self._objs.pop()), None)


class _Serializer:
    """Holds the state of a single serialize() call."""

    __slots__ = (
        "memo",
        "path",
        "meta_stack",
        "keep_request_bodies",
        "max_value_length",
        "is_vars",
        "custom_repr",
    )

    def __init__(
        self,
        keep_request_bodies: bool,
        max_value_length: "Optional[int]",
        is_vars: bool,
        custom_repr: "Optional[Callable[..., Optional[str]]]",
    ) -> None:
        self.memo = Memo()
        self.path: "List[Segment]" = []
        self.meta_stack: "List[Dict[str, Any]]" = []
        self.keep_request_bodies = keep_request_bodies
        self.max_value_length = max_value_length
        self.is_vars = is_vars
        self.custom_repr = custom_repr

    def _safe_repr_wrapper(self, value: "Any") -> str:
        try:
            repr_value = None
            if self.custom_repr is not None:
                repr_value = self.custom_repr(value)
            return repr_value or safe_repr(value)
        except Exception:
            return safe_repr(value)

    def _annotate(self, **meta: "Any") -> None:
        while len(self.meta_stack) <= len(self.path):
            try:
                segment = self.path[len(self.meta_stack) - 1]
                node = self.meta_stack[-1].setdefault(str(segment), {})
            except IndexError:
                node = {}

            self.meta_stack.append(node)

        self.meta_stack[-1].setdefault("", {}).update(meta)

    def _is_databag(self) -> "Optional[bool]":
        """
        A databag is any value that we need to trim.
        True for stuff like vars, request bodies, breadcrumbs and extra.

        :returns: `True` for "yes", `False` for :"no", `None` for "maybe soon".
        """
        try:
            if self.is_vars:
                return True

            is_request_body = self._is_request_body()
            if is_request_body in (True, None):
                return is_request_body

            p0 = self.path[0]
            if p0 == "breadcrumbs" and self.path[1] == "values":
                self.path[2]
                return True

            if p0 == "extra":
                return True

        except IndexError:
            return None

        return False

    def _is_span_attribute(self) -> "Optional[bool]":
        try:
            if self.path[0] == "spans" and self.path[2] == "data":
                return True
        except IndexError:
            return None

        return False

    def _is_request_body(self) -> "Optional[bool]":
        try:
            if self.path[0] == "request" and self.path[1] == "data":
                return True
        except IndexError:
            return None

        return False

    def _serialize_node(
        self,
        obj: "Any",
        is_databag: "Optional[bool]" = None,
        is_request_body: "Optional[bool]" = None,
        should_repr_strings: "Optional[bool]" = None,
        segment: "Optional[Segment]" = None,
        remaining_breadth: "Optional[Union[int, float]]" = None,
        remaining_depth: "Optional[Union[int, float]]" = None,
    ) -> "Any":
        if segment is not None:
            self.path.append(segment)

        try:
            with self.memo.memoize(obj) as result:
                if result:
                    return CYCLE_MARKER

                return self._serialize_node_impl(
                    obj,
                    is_databag=is_databag,
                    is_request_body=is_request_body,
                    should_repr_strings=should_repr_strings,
                    remaining_depth=remaining_depth,
                    remaining_breadth=remaining_breadth,
                )
        except BaseException:
            capture_internal_exception(sys.exc_info())

            if is_databag:
                return "<failed to serialize, use init(debug=True) to see error logs>"

            return None
        finally:
            if segment is not None:
                self.path.pop()
                del self.meta_stack[len(self.path) + 1 :]

    def _flatten_annotated(self, obj: "Any") -> "Any":
        if isinstance(obj, AnnotatedValue):
            self._annotate(**obj.metadata)
            obj = obj.value
        return obj

    def _serialize_node_impl(
        self,
        obj: "Any",
        is_databag: "Optional[bool]",
        is_request_body: "Optional[bool]",
        should_repr_strings: "Optional[bool]",
        remaining_depth: "Optional[Union[float, int]]",
        remaining_breadth: "Optional[Union[float, int]]",
    ) -> "Any":
        if isinstance(obj, AnnotatedValue):
            should_repr_strings = False
        if should_repr_strings is None:
            should_repr_strings = self.is_vars

        if is_databag is None:
            is_databag = self._is_databag()

        if is_request_body is None:
            is_request_body = self._is_request_body()

        if is_databag:
            if is_request_body and self.keep_request_bodies:
                remaining_depth = float("inf")
                remaining_breadth = float("inf")
            else:
                if remaining_depth is None:
                    remaining_depth = MAX_DATABAG_DEPTH
                if remaining_breadth is None:
                    remaining_breadth = MAX_DATABAG_BREADTH

        obj = self._flatten_annotated(obj)

        if remaining_depth is not None and remaining_depth <= 0:
            self._annotate(rem=[["!limit", "x"]])
            if is_databag:
                return self._flatten_annotated(
                    strip_string(
                        self._safe_repr_wrapper(obj), max_length=self.max_value_length
                    )
                )
            return None

        is_span_attribute = self._is_span_attribute()
        if (is_databag or is_span_attribute) and global_repr_processors:
            hints = {"memo": self.memo, "remaining_depth": remaining_depth}
            for processor in global_repr_processors:
                result = processor(obj, hints)
                if result is not NotImplemented:
                    return self._flatten_annotated(result)

        sentry_repr = getattr(type(obj), "__sentry_repr__", None)

        if obj is None or isinstance(obj, (bool, int, float)):
            if should_repr_strings or (
                isinstance(obj, float) and (math.isinf(obj) or math.isnan(obj))
            ):
                return self._safe_repr_wrapper(obj)
            else:
                return obj

        elif callable(sentry_repr):
            return sentry_repr(obj)

        elif isinstance(obj, datetime):
            return (
                str(format_timestamp(obj))
                if not should_repr_strings
                else self._safe_repr_wrapper(obj)
            )

        elif isinstance(obj, Mapping):
            # Copy only as many pairs as we might keep - avoids O(len(obj)) work
            # when obj is huge (e.g. a large cache captured as a frame local),
            # while still protecting against mutation during iteration.
            obj_len = len(obj)
            if isinstance(remaining_breadth, int):
                obj = dict(islice(obj.items(), remaining_breadth + 1))
            else:
                obj = dict(obj.items())

            rv_dict: "Dict[str, Any]" = {}
            i = 0

            for k, v in obj.items():
                if remaining_breadth is not None and i >= remaining_breadth:
                    self._annotate(len=obj_len)
                    break

                str_k = str(k)
                v = self._serialize_node(
                    v,
                    segment=str_k,
                    should_repr_strings=should_repr_strings,
                    is_databag=is_databag,
                    is_request_body=is_request_body,
                    remaining_depth=(
                        remaining_depth - 1 if remaining_depth is not None else None
                    ),
                    remaining_breadth=remaining_breadth,
                )
                rv_dict[str_k] = v
                i += 1

            return rv_dict

        elif not isinstance(obj, serializable_str_types) and isinstance(
            obj, tuple(sequence_types)
        ):
            rv_list = []

            for i, v in enumerate(obj):  # type: ignore[arg-type]
                if remaining_breadth is not None and i >= remaining_breadth:
                    self._annotate(len=len(obj))  # type: ignore[arg-type]
                    break

                rv_list.append(
                    self._serialize_node(
                        v,
                        segment=i,
                        should_repr_strings=should_repr_strings,
                        is_databag=is_databag,
                        is_request_body=is_request_body,
                        remaining_depth=(
                            remaining_depth - 1 if remaining_depth is not None else None
                        ),
                        remaining_breadth=remaining_breadth,
                    )
                )

            return rv_list

        if should_repr_strings:
            obj = self._safe_repr_wrapper(obj)
        else:
            if isinstance(obj, bytes) or isinstance(obj, bytearray):
                obj = obj.decode("utf-8", "replace")

            if not isinstance(obj, str):
                obj = self._safe_repr_wrapper(obj)

        is_span_description = (
            len(self.path) == 3
            and self.path[0] == "spans"
            and self.path[-1] == "description"
        )
        if is_span_description:
            return obj

        return self._flatten_annotated(
            strip_string(obj, max_length=self.max_value_length)
        )


def serialize(event: "Dict[str, Any]", **kwargs: "Any") -> "Dict[str, Any]":
    """
    A very smart serializer that takes a dict and emits a json-friendly dict.
    Currently used for serializing the final Event and also prematurely while fetching the stack
    local variables for each frame in a stacktrace.

    It works internally with 'databags' which are arbitrary data structures like Mapping, Sequence and Set.
    The algorithm itself is a recursive graph walk down the data structures it encounters.

    It has the following responsibilities:
    * Trimming databags and keeping them within MAX_DATABAG_BREADTH and MAX_DATABAG_DEPTH.
    * Calling safe_repr() on objects appropriately to keep them informative and readable in the final payload.
    * Annotating the payload with the _meta field whenever trimming happens.

    :param max_request_body_size: If set to "always", will never trim request bodies.
    :param max_value_length: The max length to strip strings to, or None to disable string truncation. Defaults to None.
    :param is_vars: If we're serializing vars early, we want to repr() things that are JSON-serializable to make their type more apparent. For example, it's useful to see the difference between a unicode-string and a bytestring when viewing a stacktrace.
    :param custom_repr: A custom repr function that runs before safe_repr on the object to be serialized. If it returns None or throws internally, we will fallback to safe_repr.

    """
    serializer = _Serializer(
        keep_request_bodies=kwargs.pop("max_request_body_size", None) == "always",
        max_value_length=kwargs.pop("max_value_length", None),
        is_vars=kwargs.pop("is_vars", False),
        custom_repr=kwargs.pop("custom_repr", None),
    )

    disable_capture_event.set(True)
    try:
        serialized_event = serializer._serialize_node(event, **kwargs)
        if (
            not serializer.is_vars
            and serializer.meta_stack
            and isinstance(serialized_event, dict)
        ):
            serialized_event["_meta"] = serializer.meta_stack[0]

        return serialized_event
    finally:
        disable_capture_event.set(False)



"""
Code used for the Queries module in Sentry
"""

from typing import TYPE_CHECKING

from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations.redis.utils import _get_safe_command
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.utils import capture_internal_exceptions

if TYPE_CHECKING:
    from typing import Any, Union

    from redis import Redis

    from sentry_sdk.integrations.redis import RedisIntegration
    from sentry_sdk.tracing import Span


def _compile_db_span_properties(
    integration: "RedisIntegration", redis_command: str, args: "tuple[Any, ...]"
) -> "dict[str, Any]":
    description = _get_db_span_description(integration, redis_command, args)

    properties = {
        "op": OP.DB_REDIS,
        "description": description,
    }

    return properties


def _get_db_span_description(
    integration: "RedisIntegration", command_name: str, args: "tuple[Any, ...]"
) -> str:
    description = command_name

    with capture_internal_exceptions():
        description = _get_safe_command(command_name, args)

    if integration.max_data_size and len(description) > integration.max_data_size:
        description = description[: integration.max_data_size - len("...")] + "..."

    return description


def _set_db_data_on_span(
    span: "Union[Span, StreamedSpan]", connection_params: "dict[str, Any]"
) -> None:
    db = connection_params.get("db")
    host = connection_params.get("host")
    port = connection_params.get("port")

    if isinstance(span, StreamedSpan):
        span.set_attribute(SPANDATA.DB_SYSTEM_NAME, "redis")
        span.set_attribute(SPANDATA.DB_DRIVER_NAME, "redis-py")

        if db is not None:
            span.set_attribute(SPANDATA.DB_NAMESPACE, str(db))

        if host is not None:
            span.set_attribute(SPANDATA.SERVER_ADDRESS, host)

        if port is not None:
            span.set_attribute(SPANDATA.SERVER_PORT, port)

    else:
        span.set_data(SPANDATA.DB_SYSTEM, "redis")
        span.set_data(SPANDATA.DB_DRIVER_NAME, "redis-py")

        if db is not None:
            span.set_data(SPANDATA.DB_NAME, str(db))

        if host is not None:
            span.set_data(SPANDATA.SERVER_ADDRESS, host)

        if port is not None:
            span.set_data(SPANDATA.SERVER_PORT, port)


def _set_db_data(
    span: "Union[Span, StreamedSpan]", redis_instance: "Redis[Any]"
) -> None:
    try:
        _set_db_data_on_span(span, redis_instance.connection_pool.connection_kwargs)
    except AttributeError:
        pass  # connections_kwargs may be missing in some cases



import sys
from functools import wraps
from typing import TYPE_CHECKING

from sentry_sdk import consts
from sentry_sdk.ai.monitoring import record_token_usage
from sentry_sdk.ai.utils import get_start_span_function, set_data_normalized
from sentry_sdk.consts import SPANDATA
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing_utils import has_span_streaming_enabled

if TYPE_CHECKING:
    from typing import Any, Callable, Iterator, Union

    from sentry_sdk.tracing import Span

import sentry_sdk
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.utils import capture_internal_exceptions, event_from_exception, reraise

try:
    from cohere import (
        ChatStreamEndEvent,
        NonStreamedChatResponse,
    )
    from cohere.base_client import BaseCohere
    from cohere.client import Client

    if TYPE_CHECKING:
        from cohere import StreamedChatResponse
except ImportError:
    raise DidNotEnable("Cohere not installed")

try:
    # cohere 5.9.3+
    from cohere import StreamEndStreamedChatResponse
except ImportError:
    from cohere import StreamedChatResponse_StreamEnd as StreamEndStreamedChatResponse


COLLECTED_CHAT_PARAMS = {
    "model": SPANDATA.AI_MODEL_ID,
    "k": SPANDATA.AI_TOP_K,
    "p": SPANDATA.AI_TOP_P,
    "seed": SPANDATA.AI_SEED,
    "frequency_penalty": SPANDATA.AI_FREQUENCY_PENALTY,
    "presence_penalty": SPANDATA.AI_PRESENCE_PENALTY,
    "raw_prompting": SPANDATA.AI_RAW_PROMPTING,
}

COLLECTED_PII_CHAT_PARAMS = {
    "tools": SPANDATA.AI_TOOLS,
    "preamble": SPANDATA.AI_PREAMBLE,
}

COLLECTED_CHAT_RESP_ATTRS = {
    "generation_id": SPANDATA.AI_GENERATION_ID,
    "is_search_required": SPANDATA.AI_SEARCH_REQUIRED,
    "finish_reason": SPANDATA.AI_FINISH_REASON,
}

COLLECTED_PII_CHAT_RESP_ATTRS = {
    "citations": SPANDATA.AI_CITATIONS,
    "documents": SPANDATA.AI_DOCUMENTS,
    "search_queries": SPANDATA.AI_SEARCH_QUERIES,
    "search_results": SPANDATA.AI_SEARCH_RESULTS,
    "tool_calls": SPANDATA.AI_TOOL_CALLS,
}


class CohereIntegration(Integration):
    identifier = "cohere"
    origin = f"auto.ai.{identifier}"

    def __init__(self: "CohereIntegration", include_prompts: bool = True) -> None:
        self.include_prompts = include_prompts

    @staticmethod
    def setup_once() -> None:
        BaseCohere.chat = _wrap_chat(BaseCohere.chat, streaming=False)
        Client.embed = _wrap_embed(Client.embed)
        BaseCohere.chat_stream = _wrap_chat(BaseCohere.chat_stream, streaming=True)


def _capture_exception(exc: "Any") -> None:
    event, hint = event_from_exception(
        exc,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": "cohere", "handled": False},
    )
    sentry_sdk.capture_event(event, hint=hint)


def _end_span(span: "Any") -> None:
    if isinstance(span, StreamedSpan):
        span.end()
    else:
        span.__exit__(None, None, None)


def _wrap_chat(f: "Callable[..., Any]", streaming: bool) -> "Callable[..., Any]":
    def collect_chat_response_fields(
        span: "Union[Span, StreamedSpan]",
        res: "NonStreamedChatResponse",
        include_pii: bool,
    ) -> None:
        if include_pii:
            if hasattr(res, "text"):
                set_data_normalized(
                    span,
                    SPANDATA.AI_RESPONSES,
                    [res.text],
                )
            for pii_attr in COLLECTED_PII_CHAT_RESP_ATTRS:
                if hasattr(res, pii_attr):
                    set_data_normalized(span, "ai." + pii_attr, getattr(res, pii_attr))

        for attr in COLLECTED_CHAT_RESP_ATTRS:
            if hasattr(res, attr):
                set_data_normalized(span, "ai." + attr, getattr(res, attr))

        if hasattr(res, "meta"):
            if hasattr(res.meta, "billed_units"):
                record_token_usage(
                    span,
                    input_tokens=res.meta.billed_units.input_tokens,
                    output_tokens=res.meta.billed_units.output_tokens,
                )
            elif hasattr(res.meta, "tokens"):
                record_token_usage(
                    span,
                    input_tokens=res.meta.tokens.input_tokens,
                    output_tokens=res.meta.tokens.output_tokens,
                )

            if hasattr(res.meta, "warnings"):
                set_data_normalized(span, SPANDATA.AI_WARNINGS, res.meta.warnings)

    @wraps(f)
    def new_chat(*args: "Any", **kwargs: "Any") -> "Any":
        integration = sentry_sdk.get_client().get_integration(CohereIntegration)
        is_span_streaming_enabled = has_span_streaming_enabled(
            sentry_sdk.get_client().options
        )

        if (
            integration is None
            or "message" not in kwargs
            or not isinstance(kwargs.get("message"), str)
        ):
            return f(*args, **kwargs)

        message = kwargs.get("message")

        if is_span_streaming_enabled:
            span = sentry_sdk.traces.start_span(
                name="cohere.client.Chat",
                attributes={
                    "sentry.op": consts.OP.COHERE_CHAT_COMPLETIONS_CREATE,
                    "sentry.origin": CohereIntegration.origin,
                },
            )
        else:
            span = get_start_span_function()(
                op=consts.OP.COHERE_CHAT_COMPLETIONS_CREATE,
                name="cohere.client.Chat",
                origin=CohereIntegration.origin,
            )
            span.__enter__()
        try:
            res = f(*args, **kwargs)
        except Exception as e:
            exc_info = sys.exc_info()
            with capture_internal_exceptions():
                _capture_exception(e)
                span.__exit__(*exc_info)

            reraise(*exc_info)

        with capture_internal_exceptions():
            if should_send_default_pii() and integration.include_prompts:
                set_data_normalized(
                    span,
                    SPANDATA.AI_INPUT_MESSAGES,
                    list(
                        map(
                            lambda x: {
                                "role": getattr(x, "role", "").lower(),
                                "content": getattr(x, "message", ""),
                            },
                            kwargs.get("chat_history", []),
                        )
                    )
                    + [{"role": "user", "content": message}],
                )
                for k, v in COLLECTED_PII_CHAT_PARAMS.items():
                    if k in kwargs:
                        set_data_normalized(span, v, kwargs[k])

            for k, v in COLLECTED_CHAT_PARAMS.items():
                if k in kwargs:
                    set_data_normalized(span, v, kwargs[k])
            set_data_normalized(span, SPANDATA.AI_STREAMING, False)

            if streaming:
                old_iterator = res

                def new_iterator() -> "Iterator[StreamedChatResponse]":
                    with capture_internal_exceptions():
                        for x in old_iterator:
                            if isinstance(x, ChatStreamEndEvent) or isinstance(
                                x, StreamEndStreamedChatResponse
                            ):
                                collect_chat_response_fields(
                                    span,
                                    x.response,
                                    include_pii=should_send_default_pii()
                                    and integration.include_prompts,
                                )
                            yield x
                    _end_span(span)

                return new_iterator()
            elif isinstance(res, NonStreamedChatResponse):
                collect_chat_response_fields(
                    span,
                    res,
                    include_pii=should_send_default_pii()
                    and integration.include_prompts,
                )
                _end_span(span)
            else:
                set_data_normalized(span, "unknown_response", True)
                _end_span(span)
        return res

    return new_chat


def _wrap_embed(f: "Callable[..., Any]") -> "Callable[..., Any]":
    @wraps(f)
    def new_embed(*args: "Any", **kwargs: "Any") -> "Any":
        integration = sentry_sdk.get_client().get_integration(CohereIntegration)
        if integration is None:
            return f(*args, **kwargs)

        is_span_streaming_enabled = has_span_streaming_enabled(
            sentry_sdk.get_client().options
        )

        if is_span_streaming_enabled:
            span_ctx = sentry_sdk.traces.start_span(
                name="Cohere Embedding Creation",
                attributes={
                    "sentry.op": consts.OP.COHERE_EMBEDDINGS_CREATE,
                    "sentry.origin": CohereIntegration.origin,
                },
            )
        else:
            span_ctx = get_start_span_function()(
                op=consts.OP.COHERE_EMBEDDINGS_CREATE,
                name="Cohere Embedding Creation",
                origin=CohereIntegration.origin,
            )

        with span_ctx as span:
            if "texts" in kwargs and (
                should_send_default_pii() and integration.include_prompts
            ):
                if isinstance(kwargs["texts"], str):
                    set_data_normalized(span, SPANDATA.AI_TEXTS, [kwargs["texts"]])
                elif (
                    isinstance(kwargs["texts"], list)
                    and len(kwargs["texts"]) > 0
                    and isinstance(kwargs["texts"][0], str)
                ):
                    set_data_normalized(
                        span, SPANDATA.AI_INPUT_MESSAGES, kwargs["texts"]
                    )

            if "model" in kwargs:
                set_data_normalized(span, SPANDATA.AI_MODEL_ID, kwargs["model"])
            try:
                res = f(*args, **kwargs)
            except Exception as e:
                exc_info = sys.exc_info()
                with capture_internal_exceptions():
                    _capture_exception(e)
                reraise(*exc_info)
            if (
                hasattr(res, "meta")
                and hasattr(res.meta, "billed_units")
                and hasattr(res.meta.billed_units, "input_tokens")
            ):
                record_token_usage(
                    span,
                    input_tokens=res.meta.billed_units.input_tokens,
                    total_tokens=res.meta.billed_units.input_tokens,
                )
            return res

    return new_embed



import os
import random
import threading
import time
import weakref
from collections import defaultdict
from datetime import datetime, timezone
from typing import TYPE_CHECKING

from sentry_sdk._batcher import Batcher
from sentry_sdk.envelope import Envelope, Item, PayloadRef
from sentry_sdk.utils import format_timestamp, serialize_attribute

if TYPE_CHECKING:
    from typing import Any, Callable, Optional

    from sentry_sdk._types import SpanJSON


class SpanBatcher(Batcher["SpanJSON"]):
    # MAX_BEFORE_FLUSH should be lower than MAX_BEFORE_DROP, so that there is
    # a bit of a buffer for spans that appear between the trigger to flush
    # and actually flushing the buffer.
    #
    # The max limits are all per trace (per bucket).
    MAX_ENVELOPE_SIZE = 1000  # spans
    MAX_BEFORE_FLUSH = 1000
    MAX_BEFORE_DROP = 2000
    MAX_BYTES_BEFORE_FLUSH = 5 * 1024 * 1024  # 5 MB

    FLUSH_WAIT_TIME = 5.0

    TYPE = "span"
    CONTENT_TYPE = "application/vnd.sentry.items.span.v2+json"

    def __init__(
        self,
        capture_func: "Callable[[Envelope], None]",
        record_lost_func: "Callable[..., None]",
    ) -> None:
        # Spans from different traces cannot be emitted in the same envelope
        # since the envelope contains a shared trace header. That's why we bucket
        # by trace_id, so that we can then send the buckets each in its own
        # envelope.
        # trace_id -> span buffer
        self._span_buffer: dict[str, list["SpanJSON"]] = defaultdict(list)
        self._running_size: dict[str, int] = defaultdict(lambda: 0)
        self._capture_func = capture_func
        self._record_lost_func = record_lost_func
        self._running = True
        self._lock = threading.Lock()
        self._active: "threading.local" = threading.local()

        self._last_full_flush: float = time.monotonic()  # drives time-based flushes
        self._flush_event = threading.Event()
        self._pending_flush: set[str] = set()  # buckets to be flushed

        self._flusher: "Optional[threading.Thread]" = None
        self._flusher_pid: "Optional[int]" = None

        # See https://github.com/getsentry/sentry-python/blob/051cc01640a29bfd64b1f1e2e3414c02f027dd1b/sentry_sdk/monitor.py#L41-L50
        if hasattr(os, "register_at_fork"):
            weak_reset = weakref.WeakMethod(self._reset_thread_state)

            def _reset_in_child() -> None:
                method = weak_reset()
                if method is not None:
                    method()

            os.register_at_fork(after_in_child=_reset_in_child)

    def _reset_thread_state(self) -> None:
        self._span_buffer = defaultdict(list)
        self._running_size = defaultdict(lambda: 0)
        self._running = True

        self._lock = threading.Lock()
        self._active = threading.local()

        self._last_full_flush = time.monotonic()
        self._flush_event = threading.Event()
        self._pending_flush = set()

        self._flusher = None
        self._flusher_pid = None

    def _flush_loop(self) -> None:
        self._active.flag = True
        while self._running:
            jitter = random.random() * self.FLUSH_WAIT_TIME * 0.1
            self._flush_event.wait(timeout=self.FLUSH_WAIT_TIME + jitter)
            self._flush_event.clear()

            self._flush(only_pending=True)

            if (
                time.monotonic() - self._last_full_flush
                >= self.FLUSH_WAIT_TIME + jitter
            ):
                self._flush()
                self._last_full_flush = time.monotonic()

    def add(self, span: "SpanJSON") -> None:
        # Bail out if the current thread is already executing batcher code.
        # This prevents deadlocks when code running inside the batcher (e.g.
        # _add_to_envelope during flush, or _flush_event.wait/set) triggers
        # a GC-emitted warning that routes back through the logging
        # integration into add().
        if getattr(self._active, "flag", False):
            return None

        self._active.flag = True

        try:
            if not self._ensure_thread() or self._flusher is None:
                return None

            with self._lock:
                size = len(self._span_buffer[span["trace_id"]])
                if size >= self.MAX_BEFORE_DROP:
                    self._record_lost_func(
                        reason="queue_overflow",
                        data_category="span",
                        quantity=1,
                    )
                    return None

                self._span_buffer[span["trace_id"]].append(span)
                self._running_size[span["trace_id"]] += self._estimate_size(span)

                if (
                    span["is_segment"] is True
                    or size + 1 >= self.MAX_BEFORE_FLUSH
                    or self._running_size[span["trace_id"]]
                    >= self.MAX_BYTES_BEFORE_FLUSH
                ):
                    self._pending_flush.add(span["trace_id"])
                    notify = True
                else:
                    notify = False

            if notify:
                self._flush_event.set()
        finally:
            self._active.flag = False

    @staticmethod
    def _estimate_size(item: "SpanJSON") -> int:
        # Rough estimate of serialized span size that's quick to compute.
        # 210 is the rough size of the payload without attributes, and then we
        # estimate the attributes separately.
        estimate = 210
        for value in (item.get("attributes") or {}).values():
            estimate += 50

            if isinstance(value, str):
                estimate += len(value)
            else:
                estimate += len(str(value))

        return estimate

    @staticmethod
    def _to_transport_format(item: "SpanJSON") -> "Any":
        res = {k: v for k, v in item.items() if k not in ("_segment_span",)}

        if item.get("attributes"):
            res["attributes"] = {
                k: serialize_attribute(v) for (k, v) in item["attributes"].items()
            }
        else:
            del res["attributes"]

        return res

    def _flush(self, only_pending: bool = False) -> None:
        with self._lock:
            if only_pending:
                buckets = list(self._pending_flush)
            else:
                # flush whole buffer, e.g. if the SDK is shutting down
                buckets = list(self._span_buffer.keys())

            self._pending_flush.clear()

            if not buckets:
                return

            envelopes = []

            for bucket_id in buckets:
                spans = self._span_buffer.get(bucket_id)
                if not spans:
                    continue

                dsc = spans[0]["_segment_span"]._dynamic_sampling_context()

                # Max per envelope is 1000, so if we happen to have more than
                # 1000 spans in one bucket, we'll need to separate them.
                for start in range(0, len(spans), self.MAX_ENVELOPE_SIZE):
                    end = min(start + self.MAX_ENVELOPE_SIZE, len(spans))

                    envelope = Envelope(
                        headers={
                            "sent_at": format_timestamp(datetime.now(timezone.utc)),
                            "trace": dsc,
                        }
                    )

                    envelope.add_item(
                        Item(
                            type=self.TYPE,
                            content_type=self.CONTENT_TYPE,
                            headers={
                                "item_count": end - start,
                            },
                            payload=PayloadRef(
                                json={
                                    "version": 2,
                                    "items": [
                                        self._to_transport_format(spans[j])
                                        for j in range(start, end)
                                    ],
                                }
                            ),
                        )
                    )

                    envelopes.append(envelope)

                del self._span_buffer[bucket_id]
                del self._running_size[bucket_id]

        for envelope in envelopes:
            self._capture_func(envelope)



from functools import partial
from typing import TYPE_CHECKING

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.integrations import DidNotEnable, Integration, _check_minimum_version
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.traces import StreamedSpan
from sentry_sdk.tracing import Span
from sentry_sdk.tracing_utils import add_http_breadcrumb, has_span_streaming_enabled
from sentry_sdk.utils import (
    capture_internal_exceptions,
    parse_url,
    parse_version,
)

if TYPE_CHECKING:
    from typing import Any, Dict, Optional, Type, Union

    from botocore.model import ServiceId

try:
    from botocore import __version__ as BOTOCORE_VERSION
    from botocore.awsrequest import AWSRequest
    from botocore.client import BaseClient
    from botocore.response import StreamingBody
except ImportError:
    raise DidNotEnable("botocore is not installed")


class Boto3Integration(Integration):
    identifier = "boto3"
    origin = f"auto.http.{identifier}"

    @staticmethod
    def setup_once() -> None:
        version = parse_version(BOTOCORE_VERSION)
        _check_minimum_version(Boto3Integration, version, "botocore")

        orig_init = BaseClient.__init__

        def sentry_patched_init(
            self: "BaseClient", *args: "Any", **kwargs: "Any"
        ) -> None:
            orig_init(self, *args, **kwargs)
            meta = self.meta
            service_id = meta.service_model.service_id
            meta.events.register(
                "request-created",
                partial(_sentry_request_created, service_id=service_id),
            )
            meta.events.register("after-call", _sentry_after_call)
            meta.events.register("after-call-error", _sentry_after_call_error)

        BaseClient.__init__ = sentry_patched_init  # type: ignore


def _sentry_request_created(
    service_id: "ServiceId", request: "AWSRequest", operation_name: str, **kwargs: "Any"
) -> None:
    description = "aws.%s.%s" % (service_id.hyphenize(), operation_name)

    client = sentry_sdk.get_client()
    if client.get_integration(Boto3Integration) is None:
        return

    parsed_url = None
    if request.url is not None:
        with capture_internal_exceptions():
            parsed_url = parse_url(request.url, sanitize=False)

    breadcrumb: "dict[str, Any]" = {}

    is_span_streaming_enabled = has_span_streaming_enabled(client.options)
    span: "Union[Span, StreamedSpan, None]" = None
    if is_span_streaming_enabled:
        if parsed_url and should_send_default_pii():
            breadcrumb.update(
                {
                    SPANDATA.URL_FULL: parsed_url.url,
                    SPANDATA.URL_QUERY: parsed_url.query,
                    SPANDATA.URL_FRAGMENT: parsed_url.fragment,
                }
            )

        if request.method is not None:
            breadcrumb[SPANDATA.HTTP_REQUEST_METHOD] = request.method

        if sentry_sdk.traces.get_current_span() is not None:
            span = sentry_sdk.traces.start_span(
                name=description,
                attributes={
                    "sentry.op": OP.HTTP_CLIENT,
                    "sentry.origin": Boto3Integration.origin,
                    SPANDATA.RPC_METHOD: f"{service_id}/{operation_name}",
                },
            )
            if parsed_url and should_send_default_pii():
                span.set_attributes(
                    {
                        SPANDATA.URL_FULL: parsed_url.url,
                        SPANDATA.URL_QUERY: parsed_url.query,
                        SPANDATA.URL_FRAGMENT: parsed_url.fragment,
                    }
                )

            if request.method is not None:
                span.set_attribute(SPANDATA.HTTP_REQUEST_METHOD, request.method)
    else:
        span = sentry_sdk.start_span(
            op=OP.HTTP_CLIENT,
            name=description,
            origin=Boto3Integration.origin,
        )

        if parsed_url:
            span.set_data("aws.request.url", parsed_url.url)
            span.set_data(SPANDATA.HTTP_QUERY, parsed_url.query)
            span.set_data(SPANDATA.HTTP_FRAGMENT, parsed_url.fragment)
            breadcrumb.update(
                {
                    "aws.request.url": parsed_url.url,
                    SPANDATA.HTTP_QUERY: parsed_url.query,
                    SPANDATA.HTTP_FRAGMENT: parsed_url.fragment,
                }
            )

        span.set_tag("aws.service_id", service_id.hyphenize())
        span.set_tag("aws.operation_name", operation_name)
        if request.method is not None:
            span.set_data(SPANDATA.HTTP_METHOD, request.method)
            breadcrumb[SPANDATA.HTTP_METHOD] = request.method

        # We do it in order for subsequent http calls/retries be
        # attached to this span.
        span.__enter__()

    add_http_breadcrumb(None, breadcrumb)

    if span is not None:
        # request.context is an open-ended data-structure
        # where we can add anything useful in request life cycle.
        request.context["_sentrysdk_span"] = span


def _sentry_after_call(
    context: "Dict[str, Any]", parsed: "Dict[str, Any]", **kwargs: "Any"
) -> None:
    span: "Optional[Union[Span, StreamedSpan]]" = context.pop("_sentrysdk_span", None)

    # Span could be absent if the integration is disabled.
    if span is None:
        return

    span.__exit__(None, None, None)

    body = parsed.get("Body")
    if not isinstance(body, StreamingBody):
        return

    streaming_span: "Union[Span, StreamedSpan]"
    if isinstance(span, StreamedSpan):
        streaming_span = sentry_sdk.traces.start_span(
            name=span.name,
            parent_span=span,
            attributes={
                "sentry.op": OP.HTTP_CLIENT_STREAM,
                "sentry.origin": Boto3Integration.origin,
            },
        )
    else:
        streaming_span = span.start_child(
            op=OP.HTTP_CLIENT_STREAM,
            name=span.description,
            origin=Boto3Integration.origin,
        )

    orig_read = body.read
    orig_close = body.close

    def sentry_streaming_body_read(*args: "Any", **kwargs: "Any") -> bytes:
        try:
            ret = orig_read(*args, **kwargs)
            if ret:
                return ret

            if isinstance(streaming_span, StreamedSpan):
                streaming_span.end()
            else:
                streaming_span.finish()
            return ret
        except Exception:
            if isinstance(streaming_span, StreamedSpan):
                streaming_span.end()
            else:
                streaming_span.finish()
            raise

    body.read = sentry_streaming_body_read  # type: ignore

    def sentry_streaming_body_close(*args: "Any", **kwargs: "Any") -> None:
        if isinstance(streaming_span, StreamedSpan):
            streaming_span.end()
        else:
            streaming_span.finish()
        orig_close(*args, **kwargs)

    body.close = sentry_streaming_body_close  # type: ignore


def _sentry_after_call_error(
    context: "Dict[str, Any]", exception: "Type[BaseException]", **kwargs: "Any"
) -> None:
    span: "Optional[Union[Span, StreamedSpan]]" = context.pop("_sentrysdk_span", None)

    # Span could be absent if the integration is disabled.
    if span is None:
        return

    span.__exit__(type(exception), exception, None)



from collections.abc import Iterable
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union

    from openai.types.chat import (
        ChatCompletionContentPartParam,
        ChatCompletionMessageParam,
        ChatCompletionSystemMessageParam,
        ChatCompletionToolUnionParam,
    )

    from sentry_sdk._types import TextPart, ToolDefinition


def _is_system_instruction(message: "ChatCompletionMessageParam") -> bool:
    return isinstance(message, dict) and message.get("role") == "system"


def _get_system_instructions(
    messages: "Iterable[ChatCompletionMessageParam]",
) -> "list[ChatCompletionMessageParam]":
    if not isinstance(messages, Iterable):
        return []

    return [message for message in messages if _is_system_instruction(message)]


def _get_text_items(
    content: "Union[str, Iterable[ChatCompletionContentPartParam]]",
) -> "list[str]":
    if isinstance(content, str):
        return [content]

    if not isinstance(content, Iterable):
        return []

    text_items = []
    for part in content:
        if isinstance(part, dict) and part.get("type") == "text":
            text = part.get("text", None)
            if text is not None:
                text_items.append(text)

    return text_items


def _transform_system_instructions(
    system_instructions: "list[ChatCompletionSystemMessageParam]",
) -> "list[TextPart]":
    instruction_text_parts: "list[TextPart]" = []

    for instruction in system_instructions:
        if not isinstance(instruction, dict):
            continue

        content = instruction.get("content")
        if content is None:
            continue

        text_parts: "list[TextPart]" = [
            {"type": "text", "content": text} for text in _get_text_items(content)
        ]
        instruction_text_parts += text_parts

    return instruction_text_parts


def _transform_tool_definitions(
    tools: "Iterable[ChatCompletionToolUnionParam]",
) -> "list[ToolDefinition]":
    """
    Transform tool definitions to the schema used by the "gen_ai.tool.definitions" attribute.
    """
    if not isinstance(tools, Iterable):
        return []

    tool_definitions = []
    for tool in tools:
        if not isinstance(tool, dict) or "type" not in tool:
            continue

        if tool["type"] == "function":
            tool_definition: "ToolDefinition" = {
                "type": "function",
            }

            if "function" not in tool:
                tool_definitions.append(tool_definition)
                continue

            if "name" in tool["function"]:
                tool_definition["name"] = tool["function"]["name"]

            if "description" in tool["function"]:
                tool_definition["description"] = tool["function"]["description"]

            if "parameters" in tool["function"]:
                tool_definition["parameters"] = tool["function"]["parameters"]

            tool_definitions.append(tool_definition)
            continue

        if tool["type"] == "custom":
            tool_definition = {
                "type": "custom",
            }

            if "custom" not in tool:
                tool_definitions.append(tool_definition)
                continue

            if "name" in tool["custom"]:
                tool_definition["name"] = tool["custom"]["name"]

            if "description" in tool["custom"]:
                tool_definition["description"] = tool["custom"]["description"]

            tool_definitions.append(tool_definition)
            continue

    return tool_definitions



from copy import deepcopy

import sentry_sdk
from sentry_sdk.consts import OP, SPANDATA
from sentry_sdk.data_collection import _apply_key_value_collection_filtering
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.integrations.asgi import SentryAsgiMiddleware
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.tracing import SOURCE_FOR_STYLE, TransactionSource
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    ensure_integration_enabled,
    event_from_exception,
    has_data_collection_enabled,
    nullcontext,
    transaction_from_function,
)

try:
    from pydantic import BaseModel
    from starlite import Request, Starlite, State  # type: ignore
    from starlite.handlers.base import BaseRouteHandler  # type: ignore
    from starlite.middleware import DefineMiddleware  # type: ignore
    from starlite.plugins.base import get_plugin_for_value  # type: ignore
    from starlite.routes.http import HTTPRoute  # type: ignore
    from starlite.utils import (  # type: ignore
        ConnectionDataExtractor,
        Ref,
        is_async_callable,
    )
except ImportError:
    raise DidNotEnable("Starlite is not installed")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Optional, Union

    from starlite import MiddlewareProtocol
    from starlite.types import (  # type: ignore
        ASGIApp,
        Hint,
        HTTPReceiveMessage,
        HTTPScope,
        Message,
        Middleware,
        Receive,
        Send,
        WebSocketReceiveMessage,
    )
    from starlite.types import (
        Scope as StarliteScope,
    )

    from sentry_sdk._types import Event


_DEFAULT_TRANSACTION_NAME = "generic Starlite request"


class StarliteIntegration(Integration):
    identifier = "starlite"
    origin = f"auto.http.{identifier}"

    @staticmethod
    def setup_once() -> None:
        patch_app_init()
        patch_middlewares()
        patch_http_route_handle()


class SentryStarliteASGIMiddleware(SentryAsgiMiddleware):
    def __init__(
        self, app: "ASGIApp", span_origin: str = StarliteIntegration.origin
    ) -> None:
        super().__init__(
            app=app,
            unsafe_context_data=False,
            transaction_style="endpoint",
            mechanism_type="asgi",
            span_origin=span_origin,
            asgi_version=3,
        )


def patch_app_init() -> None:
    """
    Replaces the Starlite class's `__init__` function in order to inject `after_exception` handlers and set the
    `SentryStarliteASGIMiddleware` as the outmost middleware in the stack.
    See:
    - https://starlite-api.github.io/starlite/usage/0-the-starlite-app/5-application-hooks/#after-exception
    - https://starlite-api.github.io/starlite/usage/7-middleware/0-middleware-intro/
    """
    old__init__ = Starlite.__init__

    @ensure_integration_enabled(StarliteIntegration, old__init__)
    def injection_wrapper(self: "Starlite", *args: "Any", **kwargs: "Any") -> None:
        after_exception = kwargs.pop("after_exception", [])
        kwargs.update(
            after_exception=[
                exception_handler,
                *(
                    after_exception
                    if isinstance(after_exception, list)
                    else [after_exception]
                ),
            ]
        )

        middleware = kwargs.get("middleware") or []
        kwargs["middleware"] = [SentryStarliteASGIMiddleware, *middleware]
        old__init__(self, *args, **kwargs)

    Starlite.__init__ = injection_wrapper


def patch_middlewares() -> None:
    old_resolve_middleware_stack = BaseRouteHandler.resolve_middleware

    @ensure_integration_enabled(StarliteIntegration, old_resolve_middleware_stack)
    def resolve_middleware_wrapper(self: "BaseRouteHandler") -> "list[Middleware]":
        return [
            enable_span_for_middleware(middleware)
            for middleware in old_resolve_middleware_stack(self)
        ]

    BaseRouteHandler.resolve_middleware = resolve_middleware_wrapper


def enable_span_for_middleware(middleware: "Middleware") -> "Middleware":
    if (
        not hasattr(middleware, "__call__")  # noqa: B004
        or middleware is SentryStarliteASGIMiddleware
    ):
        return middleware

    if isinstance(middleware, DefineMiddleware):
        old_call: "ASGIApp" = middleware.middleware.__call__
    else:
        old_call = middleware.__call__

    async def _create_span_call(
        self: "MiddlewareProtocol",
        scope: "StarliteScope",
        receive: "Receive",
        send: "Send",
    ) -> None:
        client = sentry_sdk.get_client()
        if client.get_integration(StarliteIntegration) is None:
            return await old_call(self, scope, receive, send)

        middleware_name = self.__class__.__name__
        is_span_streaming_enabled = has_span_streaming_enabled(client.options)

        def _start_middleware_span(op: str, name: str) -> "Any":
            if is_span_streaming_enabled:
                if sentry_sdk.traces.get_current_span() is None:
                    return nullcontext()
                return sentry_sdk.traces.start_span(
                    name=name,
                    attributes={
                        "sentry.op": op,
                        "sentry.origin": StarliteIntegration.origin,
                        SPANDATA.MIDDLEWARE_NAME: middleware_name,
                    },
                )
            return sentry_sdk.start_span(
                op=op,
                name=name,
                origin=StarliteIntegration.origin,
            )

        with _start_middleware_span(
            op=OP.MIDDLEWARE_STARLITE, name=middleware_name
        ) as middleware_span:
            if not is_span_streaming_enabled:
                middleware_span.set_tag("starlite.middleware_name", middleware_name)

            # Creating spans for the "receive" callback
            async def _sentry_receive(
                *args: "Any", **kwargs: "Any"
            ) -> "Union[HTTPReceiveMessage, WebSocketReceiveMessage]":
                if sentry_sdk.get_client().get_integration(StarliteIntegration) is None:
                    return await receive(*args, **kwargs)
                with _start_middleware_span(
                    op=OP.MIDDLEWARE_STARLITE_RECEIVE,
                    name=getattr(receive, "__qualname__", str(receive)),
                ) as span:
                    if not is_span_streaming_enabled:
                        span.set_tag("starlite.middleware_name", middleware_name)
                    return await receive(*args, **kwargs)

            receive_name = getattr(receive, "__name__", str(receive))
            receive_patched = receive_name == "_sentry_receive"
            new_receive = _sentry_receive if not receive_patched else receive

            # Creating spans for the "send" callback
            async def _sentry_send(message: "Message") -> None:
                if sentry_sdk.get_client().get_integration(StarliteIntegration) is None:
                    return await send(message)
                with _start_middleware_span(
                    op=OP.MIDDLEWARE_STARLITE_SEND,
                    name=getattr(send, "__qualname__", str(send)),
                ) as span:
                    if not is_span_streaming_enabled:
                        span.set_tag("starlite.middleware_name", middleware_name)
                    return await send(message)

            send_name = getattr(send, "__name__", str(send))
            send_patched = send_name == "_sentry_send"
            new_send = _sentry_send if not send_patched else send

            return await old_call(self, scope, new_receive, new_send)

    not_yet_patched = old_call.__name__ not in ["_create_span_call"]

    if not_yet_patched:
        if isinstance(middleware, DefineMiddleware):
            middleware.middleware.__call__ = _create_span_call
        else:
            middleware.__call__ = _create_span_call

    return middleware


def patch_http_route_handle() -> None:
    old_handle = HTTPRoute.handle

    async def handle_wrapper(
        self: "HTTPRoute", scope: "HTTPScope", receive: "Receive", send: "Send"
    ) -> None:
        client = sentry_sdk.get_client()
        if client.get_integration(StarliteIntegration) is None:
            return await old_handle(self, scope, receive, send)

        sentry_scope = sentry_sdk.get_isolation_scope()
        request: "Request[Any, Any]" = scope["app"].request_class(
            scope=scope, receive=receive, send=send
        )
        extracted_request_data = ConnectionDataExtractor(
            parse_body=True, parse_query=True
        )(request)
        body = extracted_request_data.pop("body")

        request_data = await body

        route_handler = scope.get("route_handler")

        func = None
        if route_handler.name is not None:
            name = route_handler.name
        elif isinstance(route_handler.fn, Ref):
            func = route_handler.fn.value
        else:
            func = route_handler.fn
        if func is not None:
            name = transaction_from_function(func)

        source = SOURCE_FOR_STYLE["endpoint"]

        if not name:
            name = _DEFAULT_TRANSACTION_NAME
            source = TransactionSource.ROUTE

        sentry_sdk.set_transaction_name(name, source)
        sentry_scope.set_transaction_name(name, source)

        def event_processor(event: "Event", _: "Hint") -> "Event":
            request_info = event.get("request", {})
            request_info["content_length"] = len(scope.get("_body", b""))
            if has_data_collection_enabled(client.options):
                cookies = _apply_key_value_collection_filtering(
                    items=extracted_request_data["cookies"],
                    behaviour=client.options["data_collection"]["cookies"],
                )
                if cookies:
                    request_info["cookies"] = cookies
            elif should_send_default_pii():
                request_info["cookies"] = extracted_request_data["cookies"]
            if request_data is not None:
                request_info["data"] = request_data

            event["request"] = deepcopy(request_info)
            return event

        sentry_scope._name = StarliteIntegration.identifier
        sentry_scope.add_event_processor(event_processor)

        return await old_handle(self, scope, receive, send)

    HTTPRoute.handle = handle_wrapper


def retrieve_user_from_scope(scope: "StarliteScope") -> "Optional[dict[str, Any]]":
    scope_user = scope.get("user")
    if not scope_user:
        return None
    if isinstance(scope_user, dict):
        return scope_user
    if isinstance(scope_user, BaseModel):
        return scope_user.dict()
    if hasattr(scope_user, "asdict"):  # dataclasses
        return scope_user.asdict()

    plugin = get_plugin_for_value(scope_user)
    if plugin and not is_async_callable(plugin.to_dict):
        return plugin.to_dict(scope_user)

    return None


@ensure_integration_enabled(StarliteIntegration)
def exception_handler(exc: Exception, scope: "StarliteScope", _: "State") -> None:
    user_info: "Optional[dict[str, Any]]" = None
    client_options = sentry_sdk.get_client().options

    if has_data_collection_enabled(client_options):
        if client_options["data_collection"]["user_info"]:
            user_info = retrieve_user_from_scope(scope)
    elif should_send_default_pii():
        user_info = retrieve_user_from_scope(scope)

    if user_info and isinstance(user_info, dict):
        sentry_scope = sentry_sdk.get_isolation_scope()
        sentry_scope.set_user(user_info)

    event, hint = event_from_exception(
        exc,
        client_options=sentry_sdk.get_client().options,
        mechanism={"type": StarliteIntegration.identifier, "handled": False},
    )

    sentry_sdk.capture_event(event, hint=hint)



import json
from typing import TypeVar

import sentry_sdk
from sentry_sdk.api import continue_trace, get_baggage, get_traceparent
from sentry_sdk.consts import OP, SPANDATA, SPANSTATUS
from sentry_sdk.integrations import DidNotEnable, Integration
from sentry_sdk.integrations._wsgi_common import request_body_within_bounds
from sentry_sdk.traces import SegmentNameSource
from sentry_sdk.tracing import (
    BAGGAGE_HEADER_NAME,
    SENTRY_TRACE_HEADER_NAME,
    TransactionSource,
)
from sentry_sdk.tracing_utils import has_span_streaming_enabled
from sentry_sdk.utils import (
    AnnotatedValue,
    capture_internal_exceptions,
    event_from_exception,
)

R = TypeVar("R")

try:
    from dramatiq.broker import Broker
    from dramatiq.errors import Retry
    from dramatiq.message import Message
    from dramatiq.middleware import Middleware, default_middleware
except ImportError:
    raise DidNotEnable("Dramatiq is not installed")

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict, Optional, Union

    from sentry_sdk._types import Event, Hint


class DramatiqIntegration(Integration):
    """
    Dramatiq integration for Sentry

    Please make sure that you call `sentry_sdk.init` *before* initializing
    your broker, as it monkey patches `Broker.__init__`.

    This integration was originally developed and maintained
    by https://github.com/jacobsvante and later donated to the Sentry
    project.
    """

    identifier = "dramatiq"
    origin = f"auto.queue.{identifier}"

    @staticmethod
    def setup_once() -> None:
        _patch_dramatiq_broker()


def _patch_dramatiq_broker() -> None:
    original_broker__init__ = Broker.__init__

    def sentry_patched_broker__init__(
        self: "Broker", *args: "Any", **kw: "Any"
    ) -> None:
        integration = sentry_sdk.get_client().get_integration(DramatiqIntegration)

        try:
            middleware = kw.pop("middleware")
        except KeyError:
            # Unfortunately Broker and StubBroker allows middleware to be
            # passed in as positional arguments, whilst RabbitmqBroker and
            # RedisBroker does not.
            if len(args) == 1:
                middleware = args[0]
                args = ()
            else:
                middleware = None

        if middleware is None:
            middleware = list(m() for m in default_middleware)
        else:
            middleware = list(middleware)

        if integration is not None:
            middleware = [m for m in middleware if not isinstance(m, SentryMiddleware)]
            middleware.insert(0, SentryMiddleware())

        kw["middleware"] = middleware
        original_broker__init__(self, *args, **kw)

    Broker.__init__ = sentry_patched_broker__init__


class SentryMiddleware(Middleware):  # type: ignore[misc]
    """
    A Dramatiq middleware that automatically captures and sends
    exceptions to Sentry.

    This is automatically added to every instantiated broker via the
    DramatiqIntegration.
    """

    SENTRY_HEADERS_NAME = "_sentry_headers"

    def before_enqueue(
        self, broker: "Broker", message: "Message[R]", delay: int
    ) -> None:
        integration = sentry_sdk.get_client().get_integration(DramatiqIntegration)
        if integration is None:
            return

        message.options[self.SENTRY_HEADERS_NAME] = {
            BAGGAGE_HEADER_NAME: get_baggage(),
            SENTRY_TRACE_HEADER_NAME: get_traceparent(),
        }

    def before_process_message(self, broker: "Broker", message: "Message[R]") -> None:
        client = sentry_sdk.get_client()
        integration = client.get_integration(DramatiqIntegration)
        if integration is None:
            return

        message._scope_manager = sentry_sdk.isolation_scope()
        scope = message._scope_manager.__enter__()
        scope.clear_breadcrumbs()
        scope.set_extra("dramatiq_message_id", message.message_id)
        scope.add_event_processor(_make_message_event_processor(message, integration))

        sentry_headers = message.options.get(self.SENTRY_HEADERS_NAME) or {}
        if "retries" in message.options:
            # start new trace in case of retrying
            sentry_headers = {}

        if has_span_streaming_enabled(client.options):
            sentry_sdk.traces.continue_trace(sentry_headers)
            span = sentry_sdk.traces.start_span(
                name=message.actor_name,
                attributes={
                    "sentry.op": OP.QUEUE_TASK_DRAMATIQ,
                    "sentry.origin": DramatiqIntegration.origin,
                    "sentry.segment.name.source": SegmentNameSource.TASK.value,
                    SPANDATA.MESSAGING_DESTINATION_NAME: message.queue_name,
                },
                parent_span=None,
            )
            message._sentry_span_ctx = span
        else:
            transaction = continue_trace(
                sentry_headers,
                name=message.actor_name,
                op=OP.QUEUE_TASK_DRAMATIQ,
                source=TransactionSource.TASK,
                origin=DramatiqIntegration.origin,
            )
            transaction.set_status(SPANSTATUS.OK)
            sentry_sdk.start_transaction(
                transaction,
                name=message.actor_name,
                op=OP.QUEUE_TASK_DRAMATIQ,
                source=TransactionSource.TASK,
            )
            transaction.__enter__()
            transaction.set_data(
                SPANDATA.MESSAGING_DESTINATION_NAME, message.queue_name
            )
            message._sentry_span_ctx = transaction

    def after_process_message(
        self,
        broker: "Broker",
        message: "Message[R]",
        *,
        result: "Optional[Any]" = None,
        exception: "Optional[Exception]" = None,
    ) -> None:
        integration = sentry_sdk.get_client().get_integration(DramatiqIntegration)
        if integration is None:
            return

        actor = broker.get_actor(message.actor_name)
        throws = message.options.get("throws") or actor.options.get("throws")

        scope_manager = message._scope_manager
        span_ctx = getattr(message, "_sentry_span_ctx", None)
        if span_ctx is None:
            return None

        is_event_capture_required = (
            exception is not None
            and not (throws and isinstance(exception, throws))
            and not isinstance(exception, Retry)
        )
        if not is_event_capture_required:
            # normal transaction finish
            span_ctx.__exit__(None, None, None)
            scope_manager.__exit__(None, None, None)
            return

        event, hint = event_from_exception(
            exception,  # type: ignore[arg-type]
            client_options=sentry_sdk.get_client().options,
            mechanism={
                "type": DramatiqIntegration.identifier,
                "handled": False,
            },
        )
        sentry_sdk.capture_event(event, hint=hint)
        # transaction error
        span_ctx.__exit__(type(exception), exception, None)
        scope_manager.__exit__(type(exception), exception, None)

    after_skip_message = after_process_message


def _make_message_event_processor(
    message: "Message[R]", integration: "DramatiqIntegration"
) -> "Callable[[Event, Hint], Optional[Event]]":
    def inner(event: "Event", hint: "Hint") -> "Optional[Event]":
        with capture_internal_exceptions():
            DramatiqMessageExtractor(message).extract_into_event(event)

        return event

    return inner


class DramatiqMessageExtractor:
    def __init__(self, message: "Message[R]") -> None:
        self.message_data = dict(message.asdict())

    def content_length(self) -> int:
        return len(json.dumps(self.message_data))

    def extract_into_event(self, event: "Event") -> None:
        client = sentry_sdk.get_client()
        if not client.is_active():
            return

        contexts = event.setdefault("contexts", {})
        request_info = contexts.setdefault("dramatiq", {})
        request_info["type"] = "dramatiq"

        data: "Optional[Union[AnnotatedValue, Dict[str, Any]]]" = None
        if not request_body_within_bounds(client, self.content_length()):
            data = AnnotatedValue.removed_because_over_size_limit()
        else:
            data = self.message_data

        request_info["data"] = data



import json
from copy import deepcopy

import sentry_sdk
from sentry_sdk._types import SENSITIVE_DATA_SUBSTITUTE
from sentry_sdk.data_collection import _apply_key_value_collection_filtering
from sentry_sdk.scope import should_send_default_pii
from sentry_sdk.utils import AnnotatedValue, has_data_collection_enabled, logger

try:
    from django.http.request import RawPostDataException

    _RAW_DATA_EXCEPTIONS = (RawPostDataException, ValueError)
except ImportError:
    RawPostDataException = None
    _RAW_DATA_EXCEPTIONS = (ValueError,)

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict, Mapping, MutableMapping, Optional, Union

    from sentry_sdk._types import Event, HttpStatusCodeRange


SENSITIVE_ENV_KEYS = (
    "REMOTE_ADDR",
    "HTTP_X_FORWARDED_FOR",
    "HTTP_SET_COOKIE",
    "HTTP_COOKIE",
    "HTTP_AUTHORIZATION",
    "HTTP_PROXY_AUTHORIZATION",
    "HTTP_X_API_KEY",
    "HTTP_X_FORWARDED_FOR",
    "HTTP_X_REAL_IP",
)

SENSITIVE_HEADERS = tuple(
    x[len("HTTP_") :] for x in SENSITIVE_ENV_KEYS if x.startswith("HTTP_")
)

DEFAULT_HTTP_METHODS_TO_CAPTURE = (
    "CONNECT",
    "DELETE",
    "GET",
    # "HEAD",  # do not capture HEAD requests by default
    # "OPTIONS",  # do not capture OPTIONS requests by default
    "PATCH",
    "POST",
    "PUT",
    "TRACE",
)


def request_body_within_bounds(
    client: "Optional[sentry_sdk.client.BaseClient]", content_length: int
) -> bool:
    if client is None:
        return False

    bodies = client.options["max_request_body_size"]
    return not (
        bodies == "never"
        or (bodies == "small" and content_length > 10**3)
        or (bodies == "medium" and content_length > 10**4)
    )


class RequestExtractor:
    """
    Base class for request extraction.
    """

    # It does not make sense to make this class an ABC because it is not used
    # for typing, only so that child classes can inherit common methods from
    # it. Only some child classes implement all methods that raise
    # NotImplementedError in this class.

    def __init__(self, request: "Any") -> None:
        self.request = request

    def extract_into_event(self, event: "Event") -> None:
        client = sentry_sdk.get_client()
        if not client.is_active():
            return

        data: "Optional[Union[AnnotatedValue, Dict[str, Any]]]" = None

        content_length = self.content_length()
        request_info = event.get("request", {})

        # Prior to data collection being implemented we unconditionally attached
        # the request body, which is why we default to True here.
        attach_request_body = True

        if has_data_collection_enabled(client.options):
            cookies = _apply_key_value_collection_filtering(
                items=dict(self.cookies()),
                behaviour=client.options["data_collection"]["cookies"],
            )
            if cookies:
                request_info["cookies"] = cookies

            attach_request_body = (
                "incoming_request" in client.options["data_collection"]["http_bodies"]
            )
        elif should_send_default_pii():
            request_info["cookies"] = dict(self.cookies())

        if attach_request_body:
            if not request_body_within_bounds(client, content_length):
                data = AnnotatedValue.removed_because_over_size_limit()
            else:
                # First read the raw body data
                # It is important to read this first because if it is Django
                # it will cache the body and then we can read the cached version
                # again in parsed_body() (or json() or wherever).
                raw_data = None
                try:
                    raw_data = self.raw_data()
                except _RAW_DATA_EXCEPTIONS:
                    # If DjangoRestFramework is used it already read the body for us
                    # so reading it here will fail. We can ignore this.
                    pass

                parsed_body = self.parsed_body()
                if parsed_body is not None:
                    data = parsed_body
                elif raw_data:
                    data = AnnotatedValue.removed_because_raw_data()
                else:
                    data = None

        if data is not None:
            request_info["data"] = data

        event["request"] = deepcopy(request_info)

    def content_length(self) -> int:
        try:
            return int(self.env().get("CONTENT_LENGTH", 0))
        except ValueError:
            return 0

    def cookies(self) -> "MutableMapping[str, Any]":
        raise NotImplementedError()

    def raw_data(self) -> "Optional[Union[str, bytes]]":
        raise NotImplementedError()

    def form(self) -> "Optional[Dict[str, Any]]":
        raise NotImplementedError()

    def parsed_body(self) -> "Optional[Dict[str, Any]]":
        try:
            form = self.form()
        except Exception:
            form = None
        try:
            files = self.files()
        except Exception:
            files = None

        if form or files:
            data = {}
            if form:
                data = dict(form.items())
            if files:
                for key in files.keys():
                    data[key] = AnnotatedValue.removed_because_raw_data()

            return data

        return self.json()

    def is_json(self) -> bool:
        return _is_json_content_type(self.env().get("CONTENT_TYPE"))

    def json(self) -> "Optional[Any]":
        try:
            if not self.is_json():
                return None

            try:
                raw_data = self.raw_data()
            except _RAW_DATA_EXCEPTIONS:
                # The body might have already been read, in which case this will
                # fail
                raw_data = None

            if raw_data is None:
                return None

            if isinstance(raw_data, str):
                return json.loads(raw_data)
            else:
                return json.loads(raw_data.decode("utf-8"))
        except ValueError:
            pass

        return None

    def files(self) -> "Optional[Dict[str, Any]]":
        raise NotImplementedError()

    def size_of_file(self, file: "Any") -> int:
        raise NotImplementedError()

    def env(self) -> "Dict[str, Any]":
        raise NotImplementedError()


def _is_json_content_type(ct: "Optional[str]") -> bool:
    mt = (ct or "").split(";", 1)[0]
    return (
        mt == "application/json"
        or (mt.startswith("application/"))
        and mt.endswith("+json")
    )


def _filter_headers(
    headers: "Mapping[str, str]",
    use_annotated_value: bool = True,
) -> "Mapping[str, Union[AnnotatedValue, str]]":
    client_options = sentry_sdk.get_client().options

    if has_data_collection_enabled(client_options):
        data_collection_configuration = client_options["data_collection"]

        filtered = _apply_key_value_collection_filtering(
            items=headers,
            behaviour=data_collection_configuration["http_headers"]["request"],
        )

        for key in filtered:
            if isinstance(key, str) and key.lower() in ("cookie", "set-cookie"):
                filtered[key] = SENSITIVE_DATA_SUBSTITUTE

        return filtered
    else:
        if should_send_default_pii():
            return headers

        substitute: "Union[AnnotatedValue, str]" = (
            SENSITIVE_DATA_SUBSTITUTE
            if not use_annotated_value
            else AnnotatedValue.removed_because_over_size_limit()
        )

        return {
            k: (
                v
                if k.upper().replace("-", "_") not in SENSITIVE_HEADERS
                else substitute
            )
            for k, v in headers.items()
        }


def _in_http_status_code_range(
    code: object, code_ranges: "list[HttpStatusCodeRange]"
) -> bool:
    for target in code_ranges:
        if isinstance(target, int):
            if code == target:
                return True
            continue

        try:
            if code in target:
                return True
        except TypeError:
            logger.warning(
                "failed_request_status_codes has to be a list of integers or containers"
            )

    return False


class HttpCodeRangeContainer:
    """
    Wrapper to make it possible to use list[HttpStatusCodeRange] as a Container[int].
    Used for backwards compatibility with the old `failed_request_status_codes` option.
    """

    def __init__(self, code_ranges: "list[HttpStatusCodeRange]") -> None:
        self._code_ranges = code_ranges

    def __contains__(self, item: object) -> bool:
        return _in_http_status_code_range(item, self._code_ranges)
